import Pyro5.api


uri = input("Enter the uri of the restricted server: ")
restricted = Pyro5.api.Proxy(uri)
print("server is running as:", restricted.who_is_server())
print("attempting to write a file:")
try:
    restricted.write_file()
    print("???? this should fail!")
except OSError as x:
    print("ERROR (expected):", x)



"""
Definition of the various exceptions that are used in Pyro.

Pyro - Python Remote Objects.  Copyright by Irmen de Jong (irmen@razorvine.net).
"""

import sys
import linecache
import traceback
from . import config


class PyroError(Exception):
    """Generic base of all Pyro-specific errors."""
    pass


class CommunicationError(PyroError):
    """Base class for the errors related to network communication problems."""
    pass


class ConnectionClosedError(CommunicationError):
    """The connection was unexpectedly closed."""
    pass


class TimeoutError(CommunicationError):
    """
    A call could not be completed within the set timeout period,
    or the network caused a timeout.
    """
    pass


class ProtocolError(CommunicationError):
    """Pyro received a message that didn't match the active Pyro network protocol, or there was a protocol related error."""
    pass


class MessageTooLargeError(ProtocolError):
    """Pyro received a message or was trying to send a message that exceeds the maximum message size as configured."""
    pass


class NamingError(PyroError):
    """There was a problem related to the name server or object names."""
    pass


class DaemonError(PyroError):
    """The Daemon encountered a problem."""
    pass


class SecurityError(PyroError):
    """A security related error occurred."""
    pass


class SerializeError(ProtocolError):
    """Something went wrong while (de)serializing data."""
    pass


def get_pyro_traceback(ex_type=None, ex_value=None, ex_tb=None):
    """Returns a list of strings that form the traceback information of a
    Pyro exception. Any remote Pyro exception information is included.
    Traceback information is automatically obtained via ``sys.exc_info()`` if
    you do not supply the objects yourself."""

    def formatRemoteTraceback(remote_tb_lines):
        result = [" +--- This exception occured remotely (Pyro) - Remote traceback:"]
        for line in remote_tb_lines:
            if line.endswith("\n"):
                line = line[:-1]
            lines = line.split("\n")
            for line2 in lines:
                result.append("\n | ")
                result.append(line2)
        result.append("\n +--- End of remote traceback\n")
        return result

    try:
        if ex_type is not None and ex_value is None and ex_tb is None:
            if type(ex_type) is not type:
                raise TypeError("invalid argument: ex_type should be an exception type, or just supply no arguments at all")
        if ex_type is None and ex_tb is None:
            ex_type, ex_value, ex_tb = sys.exc_info()

        remote_tb = getattr(ex_value, "_pyroTraceback", None)
        local_tb = format_traceback(ex_type, ex_value, ex_tb, config.DETAILED_TRACEBACK)
        if remote_tb:
            remote_tb = formatRemoteTraceback(remote_tb)
            return local_tb + remote_tb
        else:
            # hmm. no remote tb info, return just the local tb.
            return local_tb
    finally:
        # clean up cycle to traceback, to allow proper GC
        del ex_type, ex_value, ex_tb


def format_traceback(ex_type=None, ex_value=None, ex_tb=None, detailed=False):
    """Formats an exception traceback. If you ask for detailed formatting,
    the result will contain info on the variables in each stack frame.
    You don't have to provide the exception info objects, if you omit them,
    this function will obtain them itself using ``sys.exc_info()``."""
    if ex_type is None and ex_tb is None:
        ex_type, ex_value, ex_tb = sys.exc_info()
    if detailed:
        def makeStrValue(value):
            try:
                return repr(value)
            except Exception:
                try:
                    return str(value)
                except Exception:
                    return "<ERROR>"

        try:
            result = ["-" * 52 + "\n",
                      " EXCEPTION %s: %s\n" % (ex_type, ex_value),
                      " Extended stacktrace follows (most recent call last)\n"]
            skipLocals = True  # don't print the locals of the very first stack frame
            while ex_tb:
                frame = ex_tb.tb_frame
                sourceFileName = frame.f_code.co_filename
                if "self" in frame.f_locals:
                    location = "%s.%s" % (frame.f_locals["self"].__class__.__name__, frame.f_code.co_name)
                else:
                    location = frame.f_code.co_name
                result.append("-" * 52 + "\n")
                result.append("File \"%s\", line %d, in %s\n" % (sourceFileName, ex_tb.tb_lineno, location))
                result.append("Source code:\n")
                result.append("    " + linecache.getline(sourceFileName, ex_tb.tb_lineno).strip() + "\n")
                if not skipLocals:
                    names = set()
                    names.update(getattr(frame.f_code, "co_varnames", ()))
                    names.update(getattr(frame.f_code, "co_names", ()))
                    names.update(getattr(frame.f_code, "co_cellvars", ()))
                    names.update(getattr(frame.f_code, "co_freevars", ()))
                    result.append("Local values:\n")
                    for name2 in sorted(names):
                        if name2 in frame.f_locals:
                            value = frame.f_locals[name2]
                            result.append("    %s = %s\n" % (name2, makeStrValue(value)))
                            if name2 == "self":
                                # print the local variables of the class instance
                                for name3, value in vars(value).items():
                                    result.append("        self.%s = %s\n" % (name3, makeStrValue(value)))
                skipLocals = False
                ex_tb = ex_tb.tb_next
            result.append("-" * 52 + "\n")
            result.append(" EXCEPTION %s: %s\n" % (ex_type, ex_value))
            result.append("-" * 52 + "\n")
            return result
        except Exception:
            return ["-" * 52 + "\nError building extended traceback!!! :\n",
                    "".join(traceback.format_exception(*sys.exc_info())) + '-' * 52 + '\n',
                    "Original Exception follows:\n",
                    "".join(traceback.format_exception(ex_type, ex_value, ex_tb))]
    else:
        # default traceback format.
        return traceback.format_exception(ex_type, ex_value, ex_tb)


def excepthook(ex_type, ex_value, ex_tb):
    """An exception hook you can use for ``sys.excepthook``, to automatically print remote Pyro tracebacks"""
    tb = "".join(get_pyro_traceback(ex_type, ex_value, ex_tb))
    sys.stderr.write(tb)



"""
Deals with the context variables of a Pyro call.

Pyro - Python Remote Objects.  Copyright by Irmen de Jong (irmen@razorvine.net).
"""

import threading
from . import errors


class _CallContext(threading.local):
    """call context thread local"""
    def __init__(self):
        # per-thread initialization
        self.client = None
        self.client_sock_addr = None
        self.seq = 0
        self.msg_flags = 0
        self.serializer_id = 0
        self.annotations = {}
        self.response_annotations = {}
        self.correlation_id = None

    def to_global(self):
        return dict(self.__dict__)

    def from_global(self, values):
        self.client = values["client"]
        self.seq = values["seq"]
        self.msg_flags = values["msg_flags"]
        self.serializer_id = values["serializer_id"]
        self.annotations = values["annotations"]
        self.response_annotations = values["response_annotations"]
        self.correlation_id = values["correlation_id"]
        self.client_sock_addr = values["client_sock_addr"]

    def track_resource(self, resource):
        """keep a weak reference to the resource to be tracked for this connection"""
        if self.client:
            self.client.tracked_resources.add(resource)
        else:
            raise errors.PyroError("cannot track resource on a connectionless call")

    def untrack_resource(self, resource):
        """no longer track the resource for this connection"""
        if self.client:
            self.client.tracked_resources.discard(resource)
        else:
            raise errors.PyroError("cannot untrack resource on a connectionless call")


current_context = _CallContext()
"""the thread-local context object for the current Pyro call"""



"""
Tests for the data serializer.

Pyro - Python Remote Objects.  Copyright by Irmen de Jong (irmen@razorvine.net).
"""
import array
import collections
import copy
import math
import uuid
import pytest
import Pyro5.errors
import Pyro5.core
import Pyro5.client
import Pyro5.server
import Pyro5.serializers
from Pyro5 import config
from support import *


class TestSerpentSerializer:
    serializer = Pyro5.serializers.serializers["serpent"]

    def testSourceByteTypes(self):
        call_ser = self.serializer.dumpsCall("object", "method", [1, 2, 3], {"kwarg": 42})
        ser = self.serializer.dumps([4, 5, 6])
        _, _, vargs, _ = self.serializer.loadsCall(bytearray(call_ser))
        assert vargs == [1, 2, 3]
        d = self.serializer.loads(bytearray(ser))
        assert d == [4, 5, 6]
        _, _, vargs, _ = self.serializer.loadsCall(memoryview(call_ser))
        assert vargs == [1, 2, 3]
        d = self.serializer.loads(memoryview(ser))
        assert d == [4, 5, 6]

    def testSerializePyroTypes(self):
        uri = Pyro5.core.URI("PYRO:obj@host:9999")
        ser = self.serializer.dumps(uri)
        uri2 = self.serializer.loads(ser)
        assert isinstance(uri2, Pyro5.core.URI)
        assert uri2 == uri
        proxy = Pyro5.client.Proxy("PYRO:obj@host:9999")
        proxy._pyroHandshake = "handshake"
        ser = self.serializer.dumps(proxy)
        proxy2 = self.serializer.loads(ser)
        assert isinstance(proxy2, Pyro5.client.Proxy)
        assert proxy2 == proxy
        assert proxy2._pyroHandshake == "handshake"
        with Pyro5.server.Daemon(host="localhost", port=12345, nathost="localhost", natport=9876) as daemon:
            ser = self.serializer.dumps(daemon)
            daemon2 = self.serializer.loads(ser)
            assert isinstance(daemon2, Pyro5.server.Daemon)

    def testSerializeDumpsAndDumpsCall(self):
        self.serializer.dumps(uuid.uuid4())
        self.serializer.dumps(Pyro5.core.URI("PYRO:test@test:4444"))
        self.serializer.dumps(Pyro5.client.Proxy("PYRONAME:foobar"))
        self.serializer.dumpsCall("obj", "method", (1, 2, 3), {"arg1": 999})
        self.serializer.dumpsCall("obj", "method", (1, 2, array.array('i', [1, 2, 3])), {"arg1": 999})
        self.serializer.dumpsCall("obj", "method", (1, 2, array.array('i', [1, 2, 3])), {"arg1": array.array('i', [1, 2, 3])})
        self.serializer.dumpsCall("obj", "method", (1, 2, Pyro5.core.URI("PYRO:test@test:4444")), {"arg1": 999})
        self.serializer.dumpsCall("obj", "method", (1, 2, Pyro5.core.URI("PYRO:test@test:4444")), {"arg1": Pyro5.core.URI("PYRO:test@test:4444")})
        self.serializer.dumpsCall("obj", "method", (1, 2, Pyro5.client.Proxy("PYRONAME:foobar")), {"arg1": 999})
        self.serializer.dumpsCall("obj", "method", (1, 2, Pyro5.client.Proxy("PYRONAME:foobar")), {"arg1": Pyro5.client.Proxy("PYRONAME:foobar")})

    def testArrays(self):
        a1 = array.array('u', "hello")
        ser = self.serializer.dumps(a1)
        a2 = self.serializer.loads(ser)
        if type(a2) is array.array:
            assert a2 == a1
        else:
            assert a2 == "hello"
        a1 = array.array('h', [222, 333, 444, 555])
        ser = self.serializer.dumps(a1)
        a2 = self.serializer.loads(ser)
        if type(a2) is array.array:
            assert a2 == a1
        else:
            assert a2 == [222, 333, 444, 555]

    def testArrays2(self):
        a1 = array.array('u', "hello")
        ser = self.serializer.dumpsCall("obj", "method", [a1], {})
        a2 = self.serializer.loads(ser)
        a2 = a2["params"][0] if self.serializer.serializer_id == 3 else a2[2][0]      # 3=json serializer
        if type(a2) is array.array:
            assert a2 == a1
        else:
            assert a2 == "hello"
        a1 = array.array('h', [222, 333, 444, 555])
        ser = self.serializer.dumpsCall("obj", "method", [a1], {})
        a2 = self.serializer.loads(ser)
        a2 = a2["params"][0] if self.serializer.serializer_id == 3 else a2[2][0]      # 3=json serializer
        if type(a2) is array.array:
            assert a2 == a1
        else:
            assert a2 == [222, 333, 444, 555]


class TestMarshalSerializer(TestSerpentSerializer):
    serializer = Pyro5.serializers.serializers["marshal"]


class TestJsonSerializer(TestSerpentSerializer):
    serializer = Pyro5.serializers.serializers["json"]


if "msgpack" in Pyro5.serializers.serializers:
    class TestMsgpackSerializer(TestSerpentSerializer):
        serializer = Pyro5.serializers.serializers["msgpack"]


class TestSerializer2_serpent:
    SERIALIZER = "serpent"

    def setup_method(self):
        self.previous_serializer = config.SERIALIZER
        config.SERIALIZER = self.SERIALIZER
        self.serializer = Pyro5.serializers.serializers[config.SERIALIZER]

    def teardown_method(self):
        config.SERIALIZER = self.previous_serializer

    def testSerErrors(self):
        e1 = Pyro5.errors.NamingError("x")
        e1._pyroTraceback = ["this is the remote traceback"]
        orig_e = copy.copy(e1)
        e2 = Pyro5.errors.PyroError("x")
        e3 = Pyro5.errors.ProtocolError("x")
        p = self.serializer.dumps(e1)
        e = self.serializer.loads(p)
        assert isinstance(e, Pyro5.errors.NamingError)
        assert repr(e == repr(orig_e))
        assert e._pyroTraceback, "remote traceback info should be present" == ["this is the remote traceback"]
        p = self.serializer.dumps(e2)
        e = self.serializer.loads(p)
        assert isinstance(e, Pyro5.errors.PyroError)
        assert repr(e == repr(e2))
        p = self.serializer.dumps(e3)
        e = self.serializer.loads(p)
        assert isinstance(e, Pyro5.errors.ProtocolError)
        assert repr(e == repr(e3))

    def testSerializeExceptionWithAttr(self):
        ex = ZeroDivisionError("test error")
        ex._pyroTraceback = ["test traceback payload"]
        data = self.serializer.dumps(ex)
        ex2 = self.serializer.loads(data)
        assert type(ex2 == ZeroDivisionError)
        assert hasattr(ex2, "_pyroTraceback")
        assert ex2._pyroTraceback == ["test traceback payload"]

    def testSerCoreOffline(self):
        uri = Pyro5.core.URI("PYRO:9999@host.com:4444")
        p = self.serializer.dumps(uri)
        uri2 = self.serializer.loads(p)
        assert uri2 == uri
        assert uri2.protocol == "PYRO"
        assert uri2.object == "9999"
        assert uri2.location == "host.com:4444"
        assert uri2.port == 4444
        assert uri2.sockname is None

        uri = Pyro5.core.URI("PYRO:12345@./u:/tmp/socketname")
        p = self.serializer.dumps(uri)
        uri2 = self.serializer.loads(p)
        assert uri2 == uri
        assert uri2.protocol == "PYRO"
        assert uri2.object == "12345"
        assert uri2.location == "./u:/tmp/socketname"
        assert uri2.port is None
        assert uri2.sockname == "/tmp/socketname"

        proxy = Pyro5.client.Proxy("PYRO:9999@host.com:4444")
        proxy._pyroMaxRetries = 78
        assert not proxy._pyroConnection
        p = self.serializer.dumps(proxy)
        proxy2 = self.serializer.loads(p)
        assert not proxy._pyroConnection
        assert not proxy2._pyroConnection
        assert proxy._pyroUri == proxy2._pyroUri
        assert proxy2._pyroMaxRetries==0, "must be reset to defaults"

    def testNested(self):
        uri1 = Pyro5.core.URI("PYRO:1111@host.com:111")
        uri2 = Pyro5.core.URI("PYRO:2222@host.com:222")
        _ = self.serializer.dumps(uri1)
        data = [uri1, uri2]
        p = self.serializer.dumps(data)
        [u1, u2] = self.serializer.loads(p)
        assert u1 == uri1
        assert u2 == uri2

    def testSerDaemonHack(self):
        # This tests the hack that a Daemon should be serializable,
        # but only to support serializing Pyro objects.
        # The serialized form of a Daemon should be empty (and thus, useless)
        with Pyro5.server.Daemon(port=0) as daemon:
            d = self.serializer.dumps(daemon)
            d2 = self.serializer.loads(d)
            assert len(d2.__dict__) == 0, "deserialized daemon should be empty"
            assert "Pyro5.server.Daemon" in repr(d2)
            assert "unusable" in repr(d2)
            obj = MyThingFullExposed("hello")
            daemon.register(obj)
            _ = self.serializer.dumps(obj)

    def testPyroClasses(self):
        uri = Pyro5.core.URI("PYRO:object@host:4444")
        s = self.serializer.dumps(uri)
        x = self.serializer.loads(s)
        assert isinstance(x, Pyro5.core.URI)
        assert x == uri
        assert "URI" in repr(uri)
        assert str(uri == "PYRO:object@host:4444")
        uri = Pyro5.core.URI("PYRO:12345@./u:/tmp/socketname")
        s = self.serializer.dumps(uri)
        x = self.serializer.loads(s)
        assert isinstance(x, Pyro5.core.URI)
        assert x == uri
        proxy = Pyro5.client.Proxy(uri)
        proxy._pyroAttrs = set("abc")
        proxy._pyroMethods = set("def")
        proxy._pyroOneway = set("ghi")
        proxy._pyroHandshake = "apples"
        proxy._pyroMaxRetries = 78
        proxy._pyroSerializer = "serializer"
        s = self.serializer.dumps(proxy)
        x = self.serializer.loads(s)
        assert isinstance(x, Pyro5.client.Proxy)
        assert x._pyroUri == proxy._pyroUri
        assert x._pyroAttrs == set("abc")
        assert x._pyroMethods == set("def")
        assert x._pyroOneway == set("ghi")
        assert x._pyroHandshake == "apples"
        assert x._pyroSerializer == "serializer"
        assert x._pyroMaxRetries == 0, "must be reset to defaults"
        assert "Pyro5.client.Proxy" in repr(x)
        assert "Pyro5.client.Proxy" in str(x)
        daemon = Pyro5.server.Daemon()
        s = self.serializer.dumps(daemon)
        x = self.serializer.loads(s)
        assert isinstance(x, Pyro5.server.Daemon)
        assert "Pyro5.server.Daemon" in repr(x)
        assert "unusable" in repr(x)
        assert "Pyro5.server.Daemon" in str(x)
        assert "unusable" in str(x)
        wrapper = Pyro5.core._ExceptionWrapper(ZeroDivisionError("divided by zero"))
        s = self.serializer.dumps(wrapper)
        x = self.serializer.loads(s)
        assert isinstance(x, Pyro5.core._ExceptionWrapper)
        assert str(x.exception == "divided by zero")
        assert "ExceptionWrapper" in repr(x)
        assert "ExceptionWrapper" in str(x)

    def testProxySerializationCompat(self):
        proxy = Pyro5.client.Proxy("PYRO:object@host:4444")
        proxy._pyroSerializer = "serializer"
        pickle_state = proxy.__getstate__()
        assert len(pickle_state) == 6
        proxy.__setstate__(pickle_state)

    def testAutoProxyPartlyExposed(self):
        self.serializer.register_type_replacement(MyThingPartlyExposed, Pyro5.server._pyro_obj_to_auto_proxy)
        t1 = MyThingPartlyExposed("1")
        t2 = MyThingPartlyExposed("2")
        with Pyro5.server.Daemon() as d:
            d.register(t1, "thingy1")
            d.register(t2, "thingy2")
            data = [t1, ["apple", t2]]
            s = self.serializer.dumps(data)
            data = self.serializer.loads(s)
            assert data[1][0] == "apple"
            p1 = data[0]
            p2 = data[1][1]
            assert isinstance(p1, Pyro5.client.Proxy)
            assert isinstance(p2, Pyro5.client.Proxy)
            assert p1._pyroUri.object == "thingy1"
            assert p2._pyroUri.object == "thingy2"
            assert p1._pyroAttrs == {"readonly_prop1", "prop1"}
            assert p1._pyroMethods == {"exposed", "oneway"}
            assert p1._pyroOneway == {'oneway'}

    def testAutoProxyFullExposed(self):
        self.serializer.register_type_replacement(MyThingPartlyExposed, Pyro5.server._pyro_obj_to_auto_proxy)
        t1 = MyThingFullExposed("1")
        t2 = MyThingFullExposed("2")
        with Pyro5.server.Daemon() as d:
            d.register(t1, "thingy1")
            d.register(t2, "thingy2")
            data = [t1, ["apple", t2]]
            s = self.serializer.dumps(data)
            data = self.serializer.loads(s)
            assert data[1][0] == "apple"
            p1 = data[0]
            p2 = data[1][1]
            assert isinstance(p1, Pyro5.client.Proxy)
            assert isinstance(p2, Pyro5.client.Proxy)
            assert p1._pyroUri.object == "thingy1"
            assert p2._pyroUri.object == "thingy2"
            assert p1._pyroAttrs == {"prop1", "prop2", "readonly_prop1"}
            assert p1._pyroMethods == {'classmethod', 'method', 'oneway', 'staticmethod', 'exposed', "__dunder__"}
            assert p1._pyroOneway == {'oneway'}

    def testRegisterTypeReplacementSanity(self):
        self.serializer.register_type_replacement(int, lambda: None)
        with pytest.raises(ValueError):
            self.serializer.register_type_replacement(type, lambda: None)
        with pytest.raises(ValueError):
            self.serializer.register_type_replacement(42, lambda: None)

    def testCustomClassFail(self):
        o = MyThingFullExposed()
        s = self.serializer.dumps(o)
        with pytest.raises(Pyro5.errors.ProtocolError):
            _ = self.serializer.loads(s)

    def testCustomClassOk(self):
        o = MyThingPartlyExposed("test")
        Pyro5.serializers.SerializerBase.register_class_to_dict(MyThingPartlyExposed, mything_dict)
        Pyro5.serializers.SerializerBase.register_dict_to_class("CUSTOM-Mythingymabob", mything_creator)
        s = self.serializer.dumps(o)
        o2 = self.serializer.loads(s)
        assert isinstance(o2, MyThingPartlyExposed)
        assert o2.name == "test"
        # unregister the deserializer
        Pyro5.serializers.SerializerBase.unregister_dict_to_class("CUSTOM-Mythingymabob")
        with pytest.raises(Pyro5.errors.ProtocolError):
            self.serializer.loads(s)
        # unregister the serializer
        Pyro5.serializers.SerializerBase.unregister_class_to_dict(MyThingPartlyExposed)
        s = self.serializer.dumps(o)
        with pytest.raises(Pyro5.errors.SerializeError) as x:
            self.serializer.loads(s)
        msg = str(x.value)
        assert msg in ["unsupported serialized class: support.MyThingPartlyExposed",
                            "unsupported serialized class: PyroTests.support.MyThingPartlyExposed"]

    def testData(self):
        data = [42, "hello"]
        ser = self.serializer.dumps(data)
        data2 = self.serializer.loads(ser)
        assert data2 == data

    def testUnicodeData(self):
        data = "euro\u20aclowbytes\u0000\u0001\u007f\u0080\u00ff"
        ser = self.serializer.dumps(data)
        data2 = self.serializer.loads(ser)
        assert data2 == data

    def testUUID(self):
        data = uuid.uuid1()
        ser = self.serializer.dumps(data)
        data2 = self.serializer.loads(ser)
        uuid_as_str = str(data)
        assert data2==data or data2==uuid_as_str

    def testSet(self):
        data = {111, 222, 333}
        ser = self.serializer.dumps(data)
        data2 = self.serializer.loads(ser)
        assert data2 == data

    def testDeque(self):
        # serpent converts a deque into a primitive list
        deq = collections.deque([1, 2, 3, 4])
        ser = self.serializer.dumps(deq)
        data2 = self.serializer.loads(ser)
        assert data2 == [1, 2, 3, 4]

    def testCircularRefsValueError(self):
        with pytest.raises(ValueError):
            data = [42, "hello", Pyro5.client.Proxy("PYRO:dummy@dummy:4444")]
            data.append(data)
            ser = self.serializer.dumps(data)

    def testCallPlain(self):
        ser = self.serializer.dumpsCall("object", "method", ("vargs1", "vargs2"), {"kwargs": 999})
        obj, method, vargs, kwargs = self.serializer.loadsCall(ser)
        assert obj == "object"
        assert method == "method"
        assert len(vargs) == 2
        assert vargs[0] == "vargs1"
        assert vargs[1] == "vargs2"
        assert kwargs == {"kwargs": 999}

    def testCallPyroObjAsArg(self):
        uri = Pyro5.core.URI("PYRO:555@localhost:80")
        ser = self.serializer.dumpsCall("object", "method", [uri], {"thing": uri})
        obj, method, vargs, kwargs = self.serializer.loadsCall(ser)
        assert obj == "object"
        assert method == "method"
        assert vargs == [uri]
        assert kwargs == {"thing": uri}

    def testCallCustomObjAsArg(self):
        e = ZeroDivisionError("hello")
        ser = self.serializer.dumpsCall("object", "method", [e], {"thing": e})
        obj, method, vargs, kwargs = self.serializer.loadsCall(ser)
        assert obj == "object"
        assert method == "method"
        assert isinstance(vargs, list)
        assert isinstance(vargs[0], ZeroDivisionError)
        assert str(vargs[0] == "hello")
        assert isinstance(kwargs["thing"], ZeroDivisionError)
        assert str(kwargs["thing"] == "hello")

    def testSerializeException(self):
        e = ZeroDivisionError()
        d = self.serializer.dumps(e)
        e2 = self.serializer.loads(d)
        assert isinstance(e2, ZeroDivisionError)
        assert str(e2 == "")
        e = ZeroDivisionError("hello")
        d = self.serializer.dumps(e)
        e2 = self.serializer.loads(d)
        assert isinstance(e2, ZeroDivisionError)
        assert str(e2 == "hello")
        e = ZeroDivisionError("hello", 42)
        d = self.serializer.dumps(e)
        e2 = self.serializer.loads(d)
        assert isinstance(e2, ZeroDivisionError)
        assert str(e2) == "('hello', 42)"
        e.custom_attribute = 999
        ser = self.serializer.dumps(e)
        e2 = self.serializer.loads(ser)
        assert isinstance(e2, ZeroDivisionError)
        assert str(e2) == "('hello', 42)"
        assert e2.custom_attribute == 999

    def testSerializeSpecialException(self):
        assert "GeneratorExit" in Pyro5.serializers.all_exceptions
        e = GeneratorExit()
        d = self.serializer.dumps(e)
        e2 = self.serializer.loads(d)
        assert isinstance(e2, GeneratorExit)

    def testRecreateClasses(self):
        assert self.serializer.recreate_classes([1, 2, 3]) == [1, 2, 3]
        d = {"__class__": "invalid"}
        with pytest.raises(Pyro5.errors.ProtocolError):
            self.serializer.recreate_classes(d)
        d = {"__class__": "Pyro5.core.URI", "state": ['PYRO', '555', None, 'localhost', 80]}
        uri = self.serializer.recreate_classes(d)
        assert uri == Pyro5.core.URI("PYRO:555@localhost:80")
        number, uri = self.serializer.recreate_classes([1, {"uri": d}])
        assert number == 1
        assert uri["uri"] == Pyro5.core.URI("PYRO:555@localhost:80")

    def testUriSerializationWithoutSlots(self):
        u = Pyro5.core.URI("PYRO:obj@localhost:1234")
        d = self.serializer.dumps(u)
        u2 = self.serializer.loads(d)
        assert isinstance(u2, Pyro5.core.URI)
        assert str(u2) == "PYRO:obj@localhost:1234"

    def testFloatPrecision(self):
        f1 = 1482514078.54635912345
        f2 = 9876543212345.12345678987654321
        f3 = 11223344.556677889988776655e33
        floats = [f1, f2, f3]
        d = self.serializer.dumps(floats)
        v = self.serializer.loads(d)
        assert v, "float precision must not be compromised in any serializer" == floats

    def testSourceByteTypes_deserialize(self):
        call_ser = self.serializer.dumpsCall("object", "method", [1, 2, 3], {"kwarg": 42})
        ser = self.serializer.dumps([4, 5, 6])
        _, _, vargs, _ = self.serializer.loadsCall(bytearray(call_ser))
        assert vargs == [1, 2, 3]
        d = self.serializer.loads(bytearray(ser))
        assert d == [4, 5, 6]

    def testSourceByteTypes_deserialize_memoryview(self):
        call_ser = self.serializer.dumpsCall("object", "method", [1, 2, 3], {"kwarg": 42})
        ser = self.serializer.dumps([4, 5, 6])
        _, _, vargs, _ = self.serializer.loadsCall(memoryview(call_ser))
        assert vargs == [1, 2, 3]
        d = self.serializer.loads(memoryview(ser))
        assert d == [4, 5, 6]

    def testSourceByteTypes_loads(self):
        call_ser = self.serializer.dumpsCall("object", "method", [1, 2, 3], {"kwarg": 42})
        ser= self.serializer.dumps([4, 5, 6])
        _, _, vargs, _ = self.serializer.loadsCall(bytearray(call_ser))
        assert vargs == [1, 2, 3]
        d = self.serializer.loads(bytearray(ser))
        assert d == [4, 5, 6]

    def testSourceByteTypes_loads_memoryview(self):
        call_ser = self.serializer.dumpsCall("object", "method", [1, 2, 3], {"kwarg": 42})
        ser = self.serializer.dumps([4, 5, 6])
        _, _, vargs, _ = self.serializer.loadsCall(memoryview(call_ser))
        assert vargs == [1, 2, 3]
        d = self.serializer.loads(memoryview(ser))
        assert d == [4, 5, 6]

    def testSerializeDumpsAndDumpsCall(self):
        self.serializer.dumps(uuid.uuid4())
        self.serializer.dumps(Pyro5.core.URI("PYRO:test@test:4444"))
        self.serializer.dumps(Pyro5.client.Proxy("PYRONAME:foobar"))
        self.serializer.dumpsCall("obj", "method", (1, 2, 3), {"arg1": 999})
        self.serializer.dumpsCall("obj", "method", (1, 2, array.array('i', [1, 2, 3])), {"arg1": 999})
        self.serializer.dumpsCall("obj", "method", (1, 2, array.array('i', [1, 2, 3])), {"arg1": array.array('i', [1, 2, 3])})
        self.serializer.dumpsCall("obj", "method", (1, 2, Pyro5.core.URI("PYRO:test@test:4444")), {"arg1": 999})
        self.serializer.dumpsCall("obj", "method", (1, 2, Pyro5.core.URI("PYRO:test@test:4444")), {"arg1": Pyro5.core.URI("PYRO:test@test:4444")})
        self.serializer.dumpsCall("obj", "method", (1, 2, Pyro5.client.Proxy("PYRONAME:foobar")), {"arg1": 999})
        self.serializer.dumpsCall("obj", "method", (1, 2, Pyro5.client.Proxy("PYRONAME:foobar")), {"arg1": Pyro5.client.Proxy("PYRONAME:foobar")})

    def testArrays(self):
        a1 = array.array('u', "hello")
        ser = self.serializer.dumps(a1)
        a2 = self.serializer.loads(ser)
        if type(a2) is array.array:
            assert a2 == a1
        else:
            assert a2 == "hello"
        a1 = array.array('h', [222, 333, 444, 555])
        ser = self.serializer.dumps(a1)
        a2 = self.serializer.loads(ser)
        if type(a2) is array.array:
            assert a2 == a1
        else:
            assert a2 == [222, 333, 444, 555]

    def testArrays2(self):
        a1 = array.array('u', "hello")
        ser = self.serializer.dumpsCall("obj", "method", [a1], {})
        a2 = self.serializer.loads(ser)
        a2 = a2["params"][0] if self.SERIALIZER == "json" else a2[2][0]
        if type(a2) is array.array:
            assert a2 == a1
        else:
            assert a2 == "hello"
        a1 = array.array('h', [222, 333, 444, 555])
        ser = self.serializer.dumpsCall("obj", "method", [a1], {})
        a2 = self.serializer.loads(ser)
        a2 = a2["params"][0] if self.SERIALIZER == "json" else a2[2][0]
        if type(a2) is array.array:
            assert a2 == a1
        else:
            assert a2 == [222, 333, 444, 555]


class TestSerializer2_json(TestSerializer2_serpent):
    SERIALIZER = "json"

    def testSet(self):
        data = {111, 222, 333}
        ser = self.serializer.dumps(data)
        data2 = self.serializer.loads(ser)
        assert sorted(data2) == [111, 222, 333]

    def testDeque(self):
        pass    # can't serialize this in json


class TestSerializer2_marshal(TestSerializer2_serpent):
    SERIALIZER = "marshal"

    def testNested(self):
        pass    # marshall can't serialize custom objects

    def testAutoProxyPartlyExposed(self):
        pass    # marshall can't serialize custom objects

    def testAutoProxyFullExposed(self):
        pass    # marshall can't serialize custom objects

    def testRegisterTypeReplacementSanity(self):
        pass    # marshall doesn't support this feature at all

    def testDeque(self):
        pass    # marshall can't serialize custom objects


if "msgpack" in Pyro5.serializers.serializers:
    class TestSerializer2_msgpack(TestSerializer2_serpent):
        SERIALIZER = "msgpack"

        def testDeque(self):
            pass    # msgpack can't serialize this

        def testSet(self):
            data = {111, 222, 333}
            ser = self.serializer.dumps(data)
            data2 = self.serializer.loads(ser)
            assert sorted(data2) == [111, 222, 333]


class TestGenericCases:
    def testSerializersAvailable(self):
        _ = Pyro5.serializers.serializers["serpent"]
        _ = Pyro5.serializers.serializers["marshal"]
        _ = Pyro5.serializers.serializers["json"]

    def testAssignedSerializerIds(self):
        assert Pyro5.serializers.SerpentSerializer.serializer_id == 1
        assert Pyro5.serializers.MarshalSerializer.serializer_id == 2
        assert Pyro5.serializers.JsonSerializer.serializer_id == 3
        assert Pyro5.serializers.MsgpackSerializer.serializer_id == 4

    def testSerializersAvailableById(self):
        _ = Pyro5.serializers.serializers_by_id[1]  # serpent
        _ = Pyro5.serializers.serializers_by_id[2]  # marshal
        _ = Pyro5.serializers.serializers_by_id[3]  # json
        if "msgpack" in Pyro5.serializers.serializers:
            _ = Pyro5.serializers.serializers_by_id[4]  # msgpack
        assert 0 not in Pyro5.serializers.serializers_by_id
        assert 5 not in Pyro5.serializers.serializers_by_id

    def testDictClassFail(self):
        o = MyThingFullExposed("hello")
        d = Pyro5.serializers.SerializerBase.class_to_dict(o)
        assert d["name"] == "hello"
        assert d["__class__"] == "support.MyThingFullExposed"
        with pytest.raises(Pyro5.errors.ProtocolError):
            _ = Pyro5.serializers.SerializerBase.dict_to_class(d)

    def testDictException(self):
        x = ZeroDivisionError("hello", 42)
        expected = {
            "__class__": None,
            "__exception__": True,
            "args": ("hello", 42),
            "attributes": {}
        }
        expected["__class__"] = "builtins.ZeroDivisionError"
        d = Pyro5.serializers.SerializerBase.class_to_dict(x)
        assert d == expected
        x.custom_attribute = 999
        expected["attributes"] = {"custom_attribute": 999}
        d = Pyro5.serializers.SerializerBase.class_to_dict(x)
        assert d == expected

    def testDictClassOk(self):
        uri = Pyro5.core.URI("PYRO:object@host:4444")
        d = Pyro5.serializers.SerializerBase.class_to_dict(uri)
        assert d["__class__"] == "Pyro5.core.URI"
        assert "state" in d
        x = Pyro5.serializers.SerializerBase.dict_to_class(d)
        assert isinstance(x, Pyro5.core.URI)
        assert x == uri
        assert x.port == 4444
        uri = Pyro5.core.URI("PYRO:12345@./u:/tmp/socketname")
        d = Pyro5.serializers.SerializerBase.class_to_dict(uri)
        assert d["__class__"] == "Pyro5.core.URI"
        assert "state" in d
        x = Pyro5.serializers.SerializerBase.dict_to_class(d)
        assert isinstance(x, Pyro5.core.URI)
        assert x == uri
        assert x.sockname == "/tmp/socketname"

    def testCustomDictClass(self):
        o = MyThingPartlyExposed("test")
        Pyro5.serializers.SerializerBase.register_class_to_dict(MyThingPartlyExposed, mything_dict)
        Pyro5.serializers.SerializerBase.register_dict_to_class("CUSTOM-Mythingymabob", mything_creator)
        d = Pyro5.serializers.SerializerBase.class_to_dict(o)
        assert d["__class__"] == "CUSTOM-Mythingymabob"
        assert d["name"] == "test"
        x = Pyro5.serializers.SerializerBase.dict_to_class(d)
        assert isinstance(x, MyThingPartlyExposed)
        assert x.name == "test"
        # unregister the conversion functions and try again
        Pyro5.serializers.SerializerBase.unregister_class_to_dict(MyThingPartlyExposed)
        Pyro5.serializers.SerializerBase.unregister_dict_to_class("CUSTOM-Mythingymabob")
        d_orig = Pyro5.serializers.SerializerBase.class_to_dict(o)
        clsname = d_orig["__class__"]
        assert clsname.endswith("support.MyThingPartlyExposed")
        with pytest.raises(Pyro5.errors.ProtocolError):
            _ = Pyro5.serializers.SerializerBase.dict_to_class(d)

    def testExceptionNamespace(self):
        data = {'__class__': 'builtins.ZeroDivisionError',
                '__exception__': True,
                'args': ('hello', 42),
                'attributes': {"test_attribute": 99}}
        exc = Pyro5.serializers.SerializerBase.dict_to_class(data)
        assert isinstance(exc, ZeroDivisionError)
        assert repr(exc) == "ZeroDivisionError('hello', 42)"
        assert exc.test_attribute == 99

    def testExceptionNotTagged(self):
        data = {'__class__': 'builtins.ZeroDivisionError',
                'args': ('hello', 42),
                'attributes': {}}
        with pytest.raises(Pyro5.errors.SerializeError) as cm:
            _ = Pyro5.serializers.SerializerBase.dict_to_class(data)
        assert str(cm.value) == "unsupported serialized class: builtins.ZeroDivisionError"

    def testWeirdFloats(self):
        ser = Pyro5.serializers.serializers[config.SERIALIZER]
        p = ser.dumps([float("+inf"), float("-inf"), float("nan")])
        s2 = ser.loads(p)
        assert math.isinf(s2[0])
        assert math.copysign(1, s2[0]) == 1.0
        assert math.isinf(s2[1])
        assert math.copysign(1, s2[1]) == -1.0
        assert math.isnan(s2[2])


def mything_dict(obj):
    return {
        "__class__": "CUSTOM-Mythingymabob",
        "name": obj.name
    }


def mything_creator(classname, d):
    assert classname == "CUSTOM-Mythingymabob"
    assert d["__class__"] == "CUSTOM-Mythingymabob"
    return MyThingPartlyExposed(d["name"])




import time
import Pyro5.api
import Pyro5.errors


print("Autoreconnect using PYRO uri.")

# We create a proxy with a PYRO uri.
# Reconnect logic depends on the server now.
# (it needs to restart the object with the same id)
uri = input("Enter the uri that the server printed:").strip()
obj = Pyro5.api.Proxy(uri)

while True:
    print("call...")
    try:
        obj.method(42)
        print("Sleeping 1 second")
        time.sleep(1)
    except Pyro5.errors.ConnectionClosedError:  # or possibly CommunicationError
        print("Connection lost. REBINDING...")
        print("(restart the server now)")
        obj._pyroReconnect()



"""
Name Server and helper functions.

Pyro - Python Remote Objects.  Copyright by Irmen de Jong (irmen@razorvine.net).
"""

import warnings
import re
import sys
import logging
import socket
import time
import contextlib
import threading
from collections.abc import MutableMapping
try:
    import sqlite3
except ImportError:
    pass
from . import config, core, socketutil, server, errors
from .errors import NamingError, PyroError, ProtocolError


__all__ = ["start_ns_loop", "start_ns"]

log = logging.getLogger("Pyro5.naming")


class MemoryStorage(dict):
    """
    Storage implementation that is just an in-memory dict.
    (because it inherits from dict it is automatically a collections.MutableMapping)
    Stopping the nameserver will make the server instantly forget about everything.
    """
    def __init__(self, **kwargs):
        super(MemoryStorage, self).__init__(**kwargs)

    def __setitem__(self, key, value):
        uri, metadata = value
        super(MemoryStorage, self).__setitem__(key, (uri, metadata or frozenset()))

    def optimized_prefix_list(self, prefix, return_metadata=False):
        return None

    def optimized_regex_list(self, regex, return_metadata=False):
        return None

    def optimized_metadata_search(self, metadata_all=None, metadata_any=None, return_metadata=False):
        return None

    def everything(self, return_metadata=False):
        if return_metadata:
            return self.copy()
        return {name: uri for name, (uri, metadata) in self.items()}

    def remove_items(self, items):
        for item in items:
            if item in self:
                del self[item]

    def close(self):
        pass


class SqlStorage(MutableMapping):
    """
    Sqlite-based storage.
    It is just a single (name,uri) table for the names and another table for the metadata.
    Sqlite db connection objects aren't thread-safe, so a new connection is created in every method.
    """
    def __init__(self, dbfile):
        if dbfile == ":memory:":
            raise ValueError("We don't support the sqlite :memory: database type. Just use the default volatile in-memory store.")
        self.dbfile = dbfile
        with sqlite3.connect(dbfile) as db:
            db.execute("PRAGMA foreign_keys=ON")
            try:
                db.execute("SELECT COUNT(*) FROM pyro_names").fetchone()
            except sqlite3.OperationalError:
                # the table does not yet exist
                self._create_schema(db)
            else:
                # check if we need to update the existing schema
                try:
                    db.execute("SELECT COUNT(*) FROM pyro_metadata").fetchone()
                except sqlite3.OperationalError:
                    # metadata schema needs to be created and existing data migrated
                    db.execute("ALTER TABLE pyro_names RENAME TO pyro_names_old")
                    self._create_schema(db)
                    db.execute("INSERT INTO pyro_names(name, uri) SELECT name, uri FROM pyro_names_old")
                    db.execute("DROP TABLE pyro_names_old")
            db.commit()

    def _create_schema(self, db):
        db.execute("""CREATE TABLE pyro_names
            (
                id integer PRIMARY KEY,
                name nvarchar NOT NULL UNIQUE,
                uri nvarchar NOT NULL
            );""")
        db.execute("""CREATE TABLE pyro_metadata
            (
                object integer NOT NULL,
                metadata nvarchar NOT NULL,
                FOREIGN KEY(object) REFERENCES pyro_names(id)
            );""")

    def __getattr__(self, item):
        raise NotImplementedError("SqlStorage doesn't implement method/attribute '" + item + "'")

    def __getitem__(self, item):
        try:
            with sqlite3.connect(self.dbfile) as db:
                result = db.execute("SELECT id, uri FROM pyro_names WHERE name=?", (item,)).fetchone()
                if result:
                    dbid, uri = result
                    metadata = {m[0] for m in db.execute("SELECT metadata FROM pyro_metadata WHERE object=?", (dbid,)).fetchall()}
                    return uri, metadata
                else:
                    raise KeyError(item)
        except sqlite3.DatabaseError as e:
            raise NamingError("sqlite error in getitem: " + str(e))

    def __setitem__(self, key, value):
        uri, metadata = value
        try:
            with sqlite3.connect(self.dbfile) as db:
                cursor = db.cursor()
                cursor.execute("PRAGMA foreign_keys=ON")
                dbid = cursor.execute("SELECT id FROM pyro_names WHERE name=?", (key,)).fetchone()
                if dbid:
                    dbid = dbid[0]
                    cursor.execute("DELETE FROM pyro_metadata WHERE object=?", (dbid,))
                    cursor.execute("DELETE FROM pyro_names WHERE id=?", (dbid,))
                cursor.execute("INSERT INTO pyro_names(name, uri) VALUES(?,?)", (key, uri))
                if metadata:
                    object_id = cursor.lastrowid
                    for m in metadata:
                        cursor.execute("INSERT INTO pyro_metadata(object, metadata) VALUES (?,?)", (object_id, m))
                cursor.close()
                db.commit()
        except sqlite3.DatabaseError as e:
            raise NamingError("sqlite error in setitem: " + str(e))

    def __len__(self):
        try:
            with sqlite3.connect(self.dbfile) as db:
                return db.execute("SELECT count(*) FROM pyro_names").fetchone()[0]
        except sqlite3.DatabaseError as e:
            raise NamingError("sqlite error in len: " + str(e))

    def __contains__(self, item):
        try:
            with sqlite3.connect(self.dbfile) as db:
                return db.execute("SELECT EXISTS(SELECT 1 FROM pyro_names WHERE name=? LIMIT 1)", (item,)).fetchone()[0]
        except sqlite3.DatabaseError as e:
            raise NamingError("sqlite error in contains: " + str(e))

    def __delitem__(self, key):
        try:
            with sqlite3.connect(self.dbfile) as db:
                db.execute("PRAGMA foreign_keys=ON")
                dbid = db.execute("SELECT id FROM pyro_names WHERE name=?", (key,)).fetchone()
                if dbid:
                    dbid = dbid[0]
                    db.execute("DELETE FROM pyro_metadata WHERE object=?", (dbid,))
                    db.execute("DELETE FROM pyro_names WHERE id=?", (dbid,))
                db.commit()
        except sqlite3.DatabaseError as e:
            raise NamingError("sqlite error in delitem: " + str(e))

    def __iter__(self):
        try:
            with sqlite3.connect(self.dbfile) as db:
                result = db.execute("SELECT name FROM pyro_names")
                return iter([n[0] for n in result.fetchall()])
        except sqlite3.DatabaseError as e:
            raise NamingError("sqlite error in iter: " + str(e))

    def clear(self):
        try:
            with sqlite3.connect(self.dbfile) as db:
                db.execute("PRAGMA foreign_keys=ON")
                db.execute("DELETE FROM pyro_metadata")
                db.execute("DELETE FROM pyro_names")
                db.commit()
            with sqlite3.connect(self.dbfile, isolation_level=None) as db:
                db.execute("VACUUM")  # this cannot run inside a transaction.
        except sqlite3.DatabaseError as e:
            raise NamingError("sqlite error in clear: " + str(e))

    def optimized_prefix_list(self, prefix, return_metadata=False):
        try:
            with sqlite3.connect(self.dbfile) as db:
                names = {}
                if return_metadata:
                    for dbid, name, uri in db.execute("SELECT id, name, uri FROM pyro_names WHERE name LIKE ?", (prefix + '%',)).fetchall():
                        metadata = {m[0] for m in db.execute("SELECT metadata FROM pyro_metadata WHERE object=?", (dbid,)).fetchall()}
                        names[name] = uri, metadata
                else:
                    for name, uri in db.execute("SELECT name, uri FROM pyro_names WHERE name LIKE ?", (prefix + '%',)).fetchall():
                        names[name] = uri
                return names
        except sqlite3.DatabaseError as e:
            raise NamingError("sqlite error in optimized_prefix_list: " + str(e))

    def optimized_regex_list(self, regex, return_metadata=False):
        # defining a regex function isn't much better than simply regexing ourselves over the full table.
        return None

    def optimized_metadata_search(self, metadata_all=None, metadata_any=None, return_metadata=False):
        try:
            with sqlite3.connect(self.dbfile) as db:
                if metadata_any:
                    # any of the given metadata
                    params = list(metadata_any)
                    sql = "SELECT id, name, uri FROM pyro_names WHERE id IN (SELECT object FROM pyro_metadata WHERE metadata IN ({seq}))" \
                          .format(seq=",".join(['?'] * len(metadata_any)))
                else:
                    # all of the given metadata
                    params = list(metadata_all)
                    params.append(len(metadata_all))
                    sql = "SELECT id, name, uri FROM pyro_names WHERE id IN (SELECT object FROM pyro_metadata WHERE metadata IN ({seq}) " \
                          "GROUP BY object HAVING COUNT(metadata)=?)".format(seq=",".join(['?'] * len(metadata_all)))
                result = db.execute(sql, params).fetchall()
                if return_metadata:
                    names = {}
                    for dbid, name, uri in result:
                        metadata = {m[0] for m in db.execute("SELECT metadata FROM pyro_metadata WHERE object=?", (dbid,)).fetchall()}
                        names[name] = uri, metadata
                else:
                    names = {name: uri for (dbid, name, uri) in result}
                return names
        except sqlite3.DatabaseError as e:
            raise NamingError("sqlite error in optimized_metadata_search: " + str(e))

    def remove_items(self, items):
        try:
            with sqlite3.connect(self.dbfile) as db:
                db.execute("PRAGMA foreign_keys=ON")
                for item in items:
                    dbid = db.execute("SELECT id FROM pyro_names WHERE name=?", (item,)).fetchone()
                    if dbid:
                        dbid = dbid[0]
                        db.execute("DELETE FROM pyro_metadata WHERE object=?", (dbid,))
                        db.execute("DELETE FROM pyro_names WHERE id=?", (dbid,))
                db.commit()
        except sqlite3.DatabaseError as e:
            raise NamingError("sqlite error in remove_items: " + str(e))

    def everything(self, return_metadata=False):
        try:
            with sqlite3.connect(self.dbfile) as db:
                names = {}
                if return_metadata:
                    for dbid, name, uri in db.execute("SELECT id, name, uri FROM pyro_names").fetchall():
                        metadata = {m[0] for m in db.execute("SELECT metadata FROM pyro_metadata WHERE object=?", (dbid,)).fetchall()}
                        names[name] = uri, metadata
                else:
                    for name, uri in db.execute("SELECT name, uri FROM pyro_names").fetchall():
                        names[name] = uri
                return names
        except sqlite3.DatabaseError as e:
            raise NamingError("sqlite error in everything: " + str(e))

    def close(self):
        pass


@server.expose
class NameServer(object):
    """
    Pyro name server. Provides a simple flat name space to map logical object names to Pyro URIs.
    Default storage is done in an in-memory dictionary. You can provide custom storage types.
    """
    def __init__(self, storageProvider=None):
        self.storage = storageProvider
        if storageProvider is None:
            self.storage = MemoryStorage()
            log.debug("using volatile in-memory dict storage")
        self.lock = threading.RLock()

    def count(self):
        """Returns the number of name registrations."""
        return len(self.storage)

    def lookup(self, name, return_metadata=False):
        """
        Lookup the given name, returns an URI if found.
        Returns tuple (uri, metadata) if return_metadata is True.
        """
        try:
            uri, metadata = self.storage[name]
            uri = core.URI(uri)
            if return_metadata:
                return uri, set(metadata or [])
            return uri
        except KeyError:
            raise NamingError("unknown name: " + name)

    def register(self, name, uri, safe=False, metadata=None):
        """Register a name with an URI. If safe is true, name cannot be registered twice.
        The uri can be a string or an URI object. Metadata must be None, or a collection of strings."""
        if isinstance(uri, core.URI):
            uri = str(uri)
        elif not isinstance(uri, str):
            raise TypeError("only URIs or strings can be registered")
        else:
            core.URI(uri)  # check if uri is valid
        if not isinstance(name, str):
            raise TypeError("name must be a str")
        if isinstance(metadata, str):
            raise TypeError("metadata should not be a str, but another iterable (set, list, etc)")
        metadata and iter(metadata)  # validate that metadata is iterable
        with self.lock:
            if safe and name in self.storage:
                raise NamingError("name already registered: " + name)
            self.storage[name] = uri, set(metadata) if metadata else None

    def set_metadata(self, name, metadata):
        """update the metadata for an existing registration"""
        if not isinstance(name, str):
            raise TypeError("name must be a str")
        if isinstance(metadata, str):
            raise TypeError("metadata should not be a str, but another iterable (set, list, etc)")
        metadata and iter(metadata)  # validate that metadata is iterable
        with self.lock:
            try:
                uri, old_meta = self.storage[name]
                self.storage[name] = uri, set(metadata) if metadata else None
            except KeyError:
                raise NamingError("unknown name: " + name)

    def remove(self, name=None, prefix=None, regex=None):
        """Remove a registration. returns the number of items removed."""
        if name and name in self.storage and name != core.NAMESERVER_NAME:
            with self.lock:
                del self.storage[name]
            return 1
        if prefix:
            items = list(self.list(prefix=prefix).keys())
            if core.NAMESERVER_NAME in items:
                items.remove(core.NAMESERVER_NAME)
            self.storage.remove_items(items)
            return len(items)
        if regex:
            items = list(self.list(regex=regex).keys())
            if core.NAMESERVER_NAME in items:
                items.remove(core.NAMESERVER_NAME)
            self.storage.remove_items(items)
            return len(items)
        return 0

    # noinspection PyNoneFunctionAssignment
    def list(self, prefix=None, regex=None, return_metadata=False):
        """
        Retrieve the registered items as a dictionary name-to-URI. The URIs in the resulting dict
        are strings, not URI objects. You can filter by prefix or by regex.
        """
        if prefix and regex:
            raise ValueError("you can only filter on one thing at a time")
        with self.lock:
            if prefix:
                result = self.storage.optimized_prefix_list(prefix, return_metadata)
                if result is not None:
                    return result
                result = {}
                for name in self.storage:
                    if name.startswith(prefix):
                        result[name] = self.storage[name] if return_metadata else self.storage[name][0]
                return result
            elif regex:
                result = self.storage.optimized_regex_list(regex, return_metadata)
                if result is not None:
                    return result
                result = {}
                try:
                    regex = re.compile(regex)
                except re.error as x:
                    raise errors.NamingError("invalid regex: " + str(x))
                else:
                    for name in self.storage:
                        if regex.match(name):
                            result[name] = self.storage[name] if return_metadata else self.storage[name][0]
                    return result
            else:
                # just return (a copy of) everything
                return self.storage.everything(return_metadata)

    # noinspection PyNoneFunctionAssignment
    def yplookup(self, meta_all=None, meta_any=None, return_metadata=True):
        """
        Do a yellow-pages lookup for registrations that have all or any of the given metadata tags.
        By default returns the actual metadata in the result as well.
        """
        if meta_all and meta_any:
            raise ValueError("you can't use meta_all or meta_any at the same time")
        with self.lock:
            if meta_all:
                # return the entries which have all of the given metadata as (a subset of) their metadata
                if isinstance(meta_all, str):
                    raise TypeError("metadata_all should not be a str, but another iterable (set, list, etc)")
                meta_all and iter(meta_all)   # validate that metadata is iterable
                result = self.storage.optimized_metadata_search(metadata_all=meta_all, return_metadata=return_metadata)
                if result is not None:
                    return result
                meta_all = frozenset(meta_all)
                result = {}
                for name, (uri, meta) in self.storage.everything(return_metadata=True).items():
                    if meta_all.issubset(meta):
                        result[name] = (uri, meta) if return_metadata else uri
                return result
            elif meta_any:
                # return the entries which have any of the given metadata as part of their metadata
                if isinstance(meta_any, str):
                    raise TypeError("metadata_any should not be a str, but another iterable (set, list, etc)")
                meta_any and iter(meta_any)   # validate that metadata is iterable
                result = self.storage.optimized_metadata_search(metadata_any=meta_any, return_metadata=return_metadata)
                if result is not None:
                    return result
                meta_any = frozenset(meta_any)
                result = {}
                for name, (uri, meta) in self.storage.everything(return_metadata=True).items():
                    if meta_any & meta:
                        result[name] = (uri, meta) if return_metadata else uri
                return result
            else:
                return {}

    def ping(self):
        """A simple test method to check if the name server is running correctly."""
        pass


class NameServerDaemon(server.Daemon):
    """Daemon that contains the Name Server."""

    def __init__(self, host=None, port=None, unixsocket=None, nathost=None, natport=None, storage=None):
        if host is None:
            host = config.HOST
        elif not isinstance(host, str):
            host = str(host)  # take care of the occasion where host is an ipaddress.IpAddress
        if port is None:
            port = config.NS_PORT
        if nathost is None:
            nathost = config.NATHOST
        elif not isinstance(nathost, str):
            nathost = str(nathost)  # take care of the occasion where host is an ipaddress.IpAddress
        if natport is None:
            natport = config.NATPORT or None
        storage = storage or "memory"
        if storage == "memory":
            log.debug("using volatile in-memory dict storage")
            self.nameserver = NameServer(MemoryStorage())
        elif storage.startswith("sql:") and len(storage) > 4:
            sqlfile = storage[4:]
            log.debug("using persistent sql storage in file %s", sqlfile)
            self.nameserver = NameServer(SqlStorage(sqlfile))
        else:
            raise ValueError("invalid storage type '%s'" % storage)
        existing_count = self.nameserver.count()
        if existing_count > 0:
            log.debug("number of existing entries in storage: %d", existing_count)
        super(NameServerDaemon, self).__init__(host, port, unixsocket, nathost=nathost, natport=natport)
        self.register(self.nameserver, core.NAMESERVER_NAME)
        metadata = {"class:Pyro5.nameserver.NameServer"}
        self.nameserver.register(core.NAMESERVER_NAME, self.uriFor(self.nameserver), metadata=metadata)
        if config.NS_AUTOCLEAN > 0:
            if not AutoCleaner.override_autoclean_min and config.NS_AUTOCLEAN < AutoCleaner.min_autoclean_value:
                raise ValueError("NS_AUTOCLEAN cannot be smaller than " + str(AutoCleaner.min_autoclean_value))
            log.debug("autoclean enabled")
            self.cleaner_thread = AutoCleaner(self.nameserver)
            self.cleaner_thread.start()
        else:
            log.debug("autoclean not enabled")
            self.cleaner_thread = None
        log.info("nameserver daemon created")

    def close(self):
        super(NameServerDaemon, self).close()
        if self.nameserver is not None:
            self.nameserver.storage.close()
            self.nameserver = None
        if self.cleaner_thread:
            self.cleaner_thread.stop = True
            self.cleaner_thread.join()
            self.cleaner_thread = None

    def __enter__(self):
        if not self.nameserver:
            raise PyroError("cannot reuse this object")
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self.nameserver is not None:
            self.nameserver.storage.close()
        self.nameserver = None
        if self.cleaner_thread:
            self.cleaner_thread.stop = True
            self.cleaner_thread.join()
            self.cleaner_thread = None
        return super(NameServerDaemon, self).__exit__(exc_type, exc_value, traceback)

    def handleRequest(self, conn):
        try:
            return super(NameServerDaemon, self).handleRequest(conn)
        except ProtocolError as x:
            # Notify the user that a protocol error occurred.
            # This is useful for instance when a wrong serializer is used, it helps
            # a lot to immediately see what is going wrong.
            warnings.warn("Pyro protocol error occurred: " + str(x))
            raise


class AutoCleaner(threading.Thread):
    """
    Takes care of checking every registration in the name server.
    If it cannot be contacted anymore, it will be removed after ~20 seconds.
    """
    min_autoclean_value = 3
    max_unreachable_time = 20.0
    loop_delay = 2.0
    override_autoclean_min = False   # only for unit test purposes

    def __init__(self, nameserver):
        assert config.NS_AUTOCLEAN > 0
        if not self.override_autoclean_min and config.NS_AUTOCLEAN < self.min_autoclean_value:
            raise ValueError("NS_AUTOCLEAN cannot be smaller than " + str(self.min_autoclean_value))
        super(AutoCleaner, self).__init__()
        self.nameserver = nameserver
        self.stop = False
        self.daemon = True
        self.last_cleaned = time.time()
        self.unreachable = {}   # name->since when

    def run(self):
        while not self.stop:
            time.sleep(self.loop_delay)
            time_since_last_autoclean = time.time() - self.last_cleaned
            if time_since_last_autoclean < config.NS_AUTOCLEAN:
                continue
            for name, uri in self.nameserver.list().items():
                if name in (core.DAEMON_NAME, core.NAMESERVER_NAME):
                    continue
                try:
                    uri_obj = core.URI(uri)
                    timeout = config.COMMTIMEOUT or 5
                    sock = socketutil.create_socket(connect=(uri_obj.host, uri_obj.port), timeout=timeout)
                    sock.close()
                    # if we get here, the listed server is still answering on its port
                    if name in self.unreachable:
                        del self.unreachable[name]
                except socket.error:
                    if name not in self.unreachable:
                        self.unreachable[name] = time.time()
                    if time.time() - self.unreachable[name] >= self.max_unreachable_time:
                        log.info("autoclean: unregistering %s; cannot connect uri %s for %d sec", name, uri, self.max_unreachable_time)
                        self.nameserver.remove(name)
                        del self.unreachable[name]
                        continue
            self.last_cleaned = time.time()
            if self.unreachable:
                log.debug("autoclean: %d/%d names currently unreachable", len(self.unreachable), self.nameserver.count())


class BroadcastServer(object):
    class TransportServerAdapter(object):
        # this adapter is used to be able to pass the BroadcastServer to Daemon.combine() to integrate the event loops.
        def __init__(self, bcserver):
            self.sockets = [bcserver]

        def events(self, eventobjects):
            for bc in eventobjects:
                bc.processRequest()

    def __init__(self, nsUri, bchost=None, bcport=None, ipv6=False):
        self.transportServer = self.TransportServerAdapter(self)
        self.nsUri = nsUri
        if bcport is None:
            bcport = config.NS_BCPORT
        if bchost is None:
            bchost = config.NS_BCHOST
        elif not isinstance(bchost, str):
            bchost = str(bchost)  # take care of the occasion where host is an ipaddress.IpAddress
        if ":" in nsUri.host or ipv6:   # match nameserver's ip version
            bchost = bchost or "::"
            self.sock = socketutil.create_bc_socket((bchost, bcport, 0, 0), reuseaddr=config.SOCK_REUSE, timeout=2.0)
        else:
            self.sock = socketutil.create_bc_socket((bchost, bcport), reuseaddr=config.SOCK_REUSE, timeout=2.0)
        self._sockaddr = self.sock.getsockname()
        bchost = bchost or self._sockaddr[0]
        bcport = bcport or self._sockaddr[1]
        if ":" in bchost:  # ipv6
            self.locationStr = "[%s]:%d" % (bchost, bcport)
        else:
            self.locationStr = "%s:%d" % (bchost, bcport)
        log.info("ns broadcast server created on %s - %s", self.locationStr, socketutil.family_str(self.sock))
        self.running = True

    def close(self):
        log.debug("ns broadcast server closing")
        self.running = False
        with contextlib.suppress(OSError, socket.error):
            self.sock.shutdown(socket.SHUT_RDWR)
        self.sock.close()

    def getPort(self):
        return self.sock.getsockname()[1]

    def fileno(self):
        return self.sock.fileno()

    def runInThread(self):
        """Run the broadcast server loop in its own thread."""
        thread = threading.Thread(target=self.__requestLoop)
        thread.daemon = True
        thread.start()
        log.debug("broadcast server loop running in own thread")
        return thread

    def __requestLoop(self):
        while self.running:
            self.processRequest()
        log.debug("broadcast server loop terminating")

    def processRequest(self):
        with contextlib.suppress(socket.error):
            data, addr = self.sock.recvfrom(100)
            if data == b"GET_NSURI":
                responsedata = core.URI(self.nsUri)
                if responsedata.host == "0.0.0.0":
                    # replace INADDR_ANY address by the interface IP address that connects to the requesting client
                    with contextlib.suppress(socket.error):
                        interface_ip = socketutil.get_interface(addr[0]).ip
                        responsedata.host = str(interface_ip)
                log.debug("responding to broadcast request from %s: interface %s", addr[0], responsedata.host)
                responsedata = str(responsedata).encode("iso-8859-1")
                self.sock.sendto(responsedata, 0, addr)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()


def start_ns_loop(host=None, port=None, enableBroadcast=True, bchost=None, bcport=None,
                  unixsocket=None, nathost=None, natport=None, storage=None, 
                  loopCondition=lambda: True):
    """utility function that starts a new Name server and enters its requestloop."""
    daemon = NameServerDaemon(host, port, unixsocket, nathost=nathost, natport=natport, storage=storage)
    nsUri = daemon.uriFor(daemon.nameserver)
    internalUri = daemon.uriFor(daemon.nameserver, nat=False)
    bcserver = None
    if unixsocket:
        hostip = "Unix domain socket"
    else:
        hostip = daemon.sock.getsockname()[0]
        if daemon.sock.family == socket.AF_INET6:       # ipv6 doesn't have broadcast. We should probably use multicast instead...
            print("Not starting broadcast server for IPv6.")
            log.info("Not starting NS broadcast server because NS is using IPv6")
            enableBroadcast = False
        elif hostip.startswith("127.") or hostip in ("localhost", "::1"):
            print("Not starting broadcast server for localhost.")
            log.info("Not starting NS broadcast server because NS is bound to localhost")
            enableBroadcast = False
        if enableBroadcast:
            # Make sure to pass the internal uri to the broadcast responder.
            # It is almost always useless to let it return the external uri,
            # because external systems won't be able to talk to this thing anyway.
            bcserver = BroadcastServer(internalUri, bchost, bcport, ipv6=daemon.sock.family == socket.AF_INET6)
            print("Broadcast server running on %s" % bcserver.locationStr)
            bcserver.runInThread()
    existing = daemon.nameserver.count()
    if existing > 1:   # don't count our own nameserver registration
        print("Persistent store contains %d existing registrations." % existing)
    print("NS running on %s (%s)" % (daemon.locationStr, hostip))
    if daemon.natLocationStr:
        print("internal URI = %s" % internalUri)
        print("external URI = %s" % nsUri)
    else:
        print("URI = %s" % nsUri)
    sys.stdout.flush()
    try:
        daemon.requestLoop(loopCondition=loopCondition)
    finally:
        daemon.close()
        if bcserver is not None:
            bcserver.close()
    print("NS shut down.")


def start_ns(host=None, port=None, enableBroadcast=True, bchost=None, bcport=None,
             unixsocket=None, nathost=None, natport=None, storage=None):
    """utility fuction to quickly get a Name server daemon to be used in your own event loops.
    Returns (nameserverUri, nameserverDaemon, broadcastServer)."""
    daemon = NameServerDaemon(host, port, unixsocket, nathost=nathost, natport=natport, storage=storage)
    bcserver = None
    nsUri = daemon.uriFor(daemon.nameserver)
    if not unixsocket:
        hostip = daemon.sock.getsockname()[0]
        if hostip.startswith("127.") or hostip in ("localhost", "::1"):
            # not starting broadcast server for localhost.
            enableBroadcast = False
        if enableBroadcast:
            internalUri = daemon.uriFor(daemon.nameserver, nat=False)
            bcserver = BroadcastServer(internalUri, bchost, bcport, ipv6=daemon.sock.family == socket.AF_INET6)
    return nsUri, daemon, bcserver


def lookup(nameserver, name, return_metadata=False, delay_time=0):
    """
    Utility function to call nameserver.lookup,
    with the possibility of a retry loop until the asked name becomes available.
    You have to set the delay_time (or the corresponding config item)
    to the maximum number of seconds you are willing to wait.
    """
    delay_time = delay_time or config.NS_LOOKUP_DELAY
    start = time.time()
    while time.time()-start <= delay_time:
        try:
            return nameserver.lookup(name, return_metadata)
        except (errors.NamingError, errors.TimeoutError) as x:
            pass
        time.sleep(max(0.2, delay_time / 5))
    return nameserver.lookup(name, return_metadata)


def yplookup(nameserver, meta_all=None, meta_any=None, return_metadata=True, delay_time=0):
    """
    Utility function to call nameserver.yplookup,
    with the possibility of a retry loop until the asked name becomes available.
    You have to set the delay_time (or the corresponding config item)
    to the maximum number of seconds you are willing to wait.
    """
    delay_time = delay_time or config.NS_LOOKUP_DELAY
    start = time.time()
    while time.time()-start <= delay_time:
        try:
            result = nameserver.yplookup(meta_all, meta_any, return_metadata)
            if result:
                return result
        except (errors.NamingError, errors.TimeoutError) as x:
            pass
        time.sleep(max(0.2, delay_time / 5))
    return nameserver.yplookup(meta_all, meta_any, return_metadata)


def main(args=None):
    from argparse import ArgumentParser
    parser = ArgumentParser(description="Pyro name server command line launcher.")
    parser.add_argument("-n", "--host", dest="host", help="hostname to bind server on")
    parser.add_argument("-p", "--port", dest="port", type=int, help="port to bind server on (0=random)")
    parser.add_argument("-u", "--unixsocket", help="Unix domain socket name to bind server on")
    parser.add_argument("-s", "--storage", help="Storage system to use (memory, sql:file)", default="memory")
    parser.add_argument("--bchost", dest="bchost", help="hostname to bind broadcast server on (default is \"\")")
    parser.add_argument("--bcport", dest="bcport", type=int, help="port to bind broadcast server on (0=random)")
    parser.add_argument("--nathost", dest="nathost", help="external hostname in case of NAT")
    parser.add_argument("--natport", dest="natport", type=int, help="external port in case of NAT")
    parser.add_argument("-x", "--nobc", dest="enablebc", action="store_false", default=True,
                        help="don't start a broadcast server")
    options = parser.parse_args(args)
    start_ns_loop(options.host, options.port, enableBroadcast=options.enablebc,
                  bchost=options.bchost, bcport=options.bcport, unixsocket=options.unixsocket,
                  nathost=options.nathost, natport=options.natport, storage=options.storage)


if __name__ == "__main__":
    main()



import queue
from Pyro5.api import expose, behavior, serve, register_dict_to_class
from workitem import Workitem


# For 'workitem.Workitem' we register a deserialization hook to be able to get these back from Pyro
register_dict_to_class("workitem.Workitem", Workitem.from_dict)


@expose
@behavior(instance_mode="single")
class DispatcherQueue(object):
    def __init__(self):
        self.workqueue = queue.Queue()
        self.resultqueue = queue.Queue()

    def putWork(self, item):
        self.workqueue.put(item)

    def getWork(self, timeout=5):
        try:
            return self.workqueue.get(block=True, timeout=timeout)
        except queue.Empty:
            raise ValueError("no items in queue")

    def putResult(self, item):
        self.resultqueue.put(item)

    def getResult(self, timeout=5):
        try:
            return self.resultqueue.get(block=True, timeout=timeout)
        except queue.Empty:
            raise ValueError("no result available")

    def workQueueSize(self):
        return self.workqueue.qsize()

    def resultQueueSize(self):
        return self.resultqueue.qsize()


# main program

serve({
    DispatcherQueue: "example.distributed.dispatcher"
})



"""
The pyro wire protocol structures.

Pyro - Python Remote Objects.  Copyright by Irmen de Jong (irmen@razorvine.net).


Wire messages contains of a fixed size header, an optional set of annotation chunks,
and then the payload data. This class doesn't deal with the payload data:
(de)serialization and handling of that data is done elsewhere.
Annotation chunks are only parsed.

The header format is::

    0x00   4s  4   'PYRO' (message identifier)
    0x04   H   2   protocol version
    0x06   B   1   message type
    0x07   B   1   serializer id
    0x08   H   2   message flags
    0x0a   H   2   sequence number   (to identify proper request-reply sequencing)
    0x0c   I   4   data length   (max 4 Gb)
    0x10   I   4   annotations length (max 4 Gb, total of all chunks, 0 if no annotation chunks present)
    0x14   16s 16  correlation uuid
    0x24   H   2   (reserved)
    0x26   H   2   magic number 0x4dc5
    total size: 0x28 (40 bytes)

After the header, zero or more annotation chunks may follow, of the format::

    4s  4   annotation id (4 ASCII letters)
    I   4   chunk length  (max 4 Gb)
    B   x   annotation chunk databytes

After that, the actual payload data bytes follow.
"""

import struct
import logging
import zlib
import uuid
from . import config, errors
from .callcontext import current_context


log = logging.getLogger("Pyro5.protocol")

MSG_CONNECT = 1
MSG_CONNECTOK = 2
MSG_CONNECTFAIL = 3
MSG_INVOKE = 4
MSG_RESULT = 5
MSG_PING = 6
FLAGS_EXCEPTION = 1 << 0
FLAGS_COMPRESSED = 1 << 1    # compress the data, but not the annotations (if you need that, do it yourself)
FLAGS_ONEWAY = 1 << 2
FLAGS_BATCH = 1 << 3
FLAGS_ITEMSTREAMRESULT = 1 << 4
FLAGS_KEEPSERIALIZED = 1 << 5
FLAGS_CORR_ID = 1 << 6

# wire protocol version. Note that if this gets updated, Pyrolite might need an update too.
PROTOCOL_VERSION = 502
_magic_number = 0x4dc5
_header_format = '!4sHBBHHII16sHH'
_header_size = struct.calcsize(_header_format)
_magic_number_bytes = _magic_number.to_bytes(2, "big")
_protocol_version_bytes = PROTOCOL_VERSION.to_bytes(2, "big")
_empty_correlation_id = b"\0" * 16


class SendingMessage:
    """Wire protocol message that will be sent."""

    def __init__(self, msgtype, flags, seq, serializer_id, payload, annotations=None):
        self.type = msgtype
        self.seq = seq
        self.serializer_id = serializer_id
        annotations = annotations or {}
        annotations_size = sum([8 + len(v) for v in annotations.values()])
        flags &= ~FLAGS_COMPRESSED
        if config.COMPRESSION and len(payload) > 100:
            payload = zlib.compress(payload, 4)
            flags |= FLAGS_COMPRESSED
        self.flags = flags
        total_size = len(payload) + annotations_size
        if total_size > config.MAX_MESSAGE_SIZE:
            raise errors.ProtocolError("message too large ({:d}, max={:d})".format(total_size, config.MAX_MESSAGE_SIZE))
        if current_context.correlation_id:
            flags |= FLAGS_CORR_ID
            self.corr_id = current_context.correlation_id.bytes
        else:
            self.corr_id = _empty_correlation_id
        header_data = struct.pack(_header_format, b"PYRO", PROTOCOL_VERSION, msgtype, serializer_id, flags, seq,
                                  len(payload), annotations_size, self.corr_id, 0, _magic_number)
        annotation_data = []
        for k, v in annotations.items():
            if len(k) != 4:
                raise errors.ProtocolError("annotation identifier must be 4 ascii characters")
            annotation_data.append(struct.pack("!4sI", k.encode("ascii"), len(v)))
            if not isinstance(v, (bytes, bytearray, memoryview)):
                raise errors.ProtocolError("annotation data must be bytes, bytearray, or memoryview", type(v))
            annotation_data.append(v)    # note: annotations are not compressed by Pyro
        self.data = header_data + b"".join(annotation_data) + payload

    def __repr__(self):
        return "<{:s}.{:s} at 0x{:x}; type={:d} flags={:d} seq={:d} size={:d}>" \
            .format(self.__module__, self.__class__.__name__, id(self), self.type, self.flags, self.seq, len(self.data))

    @staticmethod
    def ping(pyroConnection):
        """Convenience method to send a 'ping' message and wait for the 'pong' response"""
        ping = SendingMessage(MSG_PING, 0, 0, 42, b"ping")
        pyroConnection.send(ping.data)
        recv_stub(pyroConnection, [MSG_PING])


class ReceivingMessage:
    """Wire protocol message that was received."""
    def __init__(self, header, payload=None):
        """Parses a message from the given header."""
        tag, ver, self.type, self.serializer_id, self.flags, self.seq, self.data_size, \
            self.annotations_size, self.corr_id,  _, magic = struct.unpack(_header_format, header)
        if tag != b"PYRO" or ver != PROTOCOL_VERSION or magic != _magic_number:
            raise errors.ProtocolError("invalid message or protocol version")
        if self.data_size+self.annotations_size > config.MAX_MESSAGE_SIZE:
            raise errors.ProtocolError("message too large ({:d}, max={:d})"
                                       .format(self.data_size+self.annotations_size, config.MAX_MESSAGE_SIZE))
        self.data = None
        self.annotations = {}
        if payload is not None:
            self.add_payload(payload)

    def __repr__(self):
        return "<{:s}.{:s} at 0x{:x}; type={:d} flags={:d} seq={:d} size={:d}>" \
            .format(self.__module__, self.__class__.__name__, id(self), self.type, self.flags, self.seq, len(self.data or ""))

    @staticmethod
    def validate(data):
        """Checks if the message data looks like a valid Pyro message, if not, raise an error."""
        ld = len(data)
        if ld < 4:
            raise ValueError("data must be at least 4 bytes to be able to identify")
        if not data.startswith(b"PYRO"):
            raise errors.ProtocolError("invalid data")
        if ld >= 6 and data[4:6] != _protocol_version_bytes:
            raise errors.ProtocolError("invalid protocol version: {:d}".format(int.from_bytes(data[4:6], "big")))
        if ld >= _header_size and data[38:40] != _magic_number_bytes:
            raise errors.ProtocolError("invalid magic number")

    def add_payload(self, payload):
        """Parses (annotations processing) and adds payload data to a received message."""
        assert not self.data
        if len(payload) != self.data_size + self.annotations_size:
            raise errors.ProtocolError("payload length doesn't match message header")
        if self.annotations_size:
            payload = memoryview(payload)  # avoid copying
            self.annotations = {}
            i = 0
            while i < self.annotations_size:
                annotation_id = bytes(payload[i:i+4]).decode("ascii")
                length = int.from_bytes(payload[i+4:i+8], "big")
                self.annotations[annotation_id] = payload[i+8:i+8+length]     # note: it stores a memoryview!
                i += 8 + length
            assert i == self.annotations_size
            self.data = payload[self.annotations_size:]
        else:
            self.data = payload
        if self.flags & FLAGS_COMPRESSED:
            self.data = zlib.decompress(self.data)
            self.flags &= ~FLAGS_COMPRESSED
            self.data_size = len(self.data)


def log_wiredata(logger, text, msg):
    """logs all the given properties of the wire message in the given logger"""
    num_anns = len(msg.annotations) if hasattr(msg, "annotations") else 0
    corr_bytes = bytes(msg.corr_id) if hasattr(msg, "corr_id") else _empty_correlation_id
    corr_id = uuid.UUID(bytes=corr_bytes)
    logger.debug("%s: msgtype=%d flags=0x%x ser=%d seq=%d num_annotations=%s corr_id=%s\ndata=%r" %
                 (text, msg.type, msg.flags, msg.serializer_id, msg.seq, num_anns, corr_id, bytes(msg.data)))


def recv_stub(connection, accepted_msgtypes=None):
    """
    Receives a pyro message from a given connection.
    Accepts the given message types (None=any, or pass a sequence).
    Also reads annotation chunks and the actual payload data.
    """
    # TODO decouple i/o from actual protocol logic, so that the protocol can be easily unit tested
    header = connection.recv(6)  # 'PYRO' + 2 bytes protocol version
    ReceivingMessage.validate(header)
    header += connection.recv(_header_size - 6)
    msg = ReceivingMessage(header)
    if accepted_msgtypes and msg.type not in accepted_msgtypes:
        err = "invalid msg type {:d} received (expected: {:s})".format(msg.type, ",".join(str(t) for t in accepted_msgtypes))
        log.error(err)
        exc = errors.ProtocolError(err)
        exc.pyroMsg = msg
        raise exc
    payload = connection.recv(msg.annotations_size + msg.data_size)
    msg.add_payload(payload)
    return msg



"""
Configuration settings.

Pyro - Python Remote Objects.  Copyright by Irmen de Jong (irmen@razorvine.net).
"""

import os
import platform
from . import __version__, _pyro_logfile, _pyro_loglevel


# noinspection PyAttributeOutsideInit

class Configuration:
    # Declare available config items.
    # DO NOT EDIT THESE HERE IN THIS MODULE! They are the global defaults.
    # Instead, specify them later in your own code or via environment variables.
    __slots__ = [
        "HOST", "NS_HOST", "NS_PORT", "NS_BCPORT", "NS_BCHOST", "NS_AUTOCLEAN", "NS_LOOKUP_DELAY",
        "NATHOST", "NATPORT", "COMPRESSION", "SERVERTYPE", "COMMTIMEOUT", "POLLTIMEOUT", "MAX_RETRIES",
        "SOCK_REUSE", "SOCK_NODELAY", "DETAILED_TRACEBACK", "THREADPOOL_SIZE", "THREADPOOL_SIZE_MIN",
        "MAX_MESSAGE_SIZE", "BROADCAST_ADDRS", "PREFER_IP_VERSION", "SERIALIZER", "SERPENT_BYTES_REPR",
        "ITER_STREAMING", "ITER_STREAM_LIFETIME", "ITER_STREAM_LINGER", "LOGFILE", "LOGLEVEL", "LOGWIRE",
        "SSL", "SSL_SERVERCERT", "SSL_SERVERKEY", "SSL_SERVERKEYPASSWD", "SSL_REQUIRECLIENTCERT",
        "SSL_CLIENTCERT", "SSL_CLIENTKEY", "SSL_CLIENTKEYPASSWD", "SSL_CACERTS"
    ]

    def __init__(self):
        self.reset()

    def reset(self, use_environment=True):
        """
        Reset to default config items.
        If use_environment is False, won't read environment variables settings (useful if you can't trust your env).
        """
        self.HOST = "localhost"
        self.NS_HOST = "localhost"
        self.NS_PORT = 9090
        self.NS_BCPORT = 9091
        self.NS_BCHOST = None
        self.NS_AUTOCLEAN = 0.0
        self.NS_LOOKUP_DELAY = 0.0
        self.NATHOST = None
        self.NATPORT = 0
        self.COMPRESSION = False
        self.SERVERTYPE = "thread"
        self.COMMTIMEOUT = 0.0
        self.POLLTIMEOUT = 2.0
        self.MAX_RETRIES = 0
        self.SOCK_REUSE = True  # so_reuseaddr on server sockets?
        self.SOCK_NODELAY = False  # tcp_nodelay on socket?
        self.DETAILED_TRACEBACK = False
        self.THREADPOOL_SIZE = 80
        self.THREADPOOL_SIZE_MIN = 4
        self.MAX_MESSAGE_SIZE = 1024 * 1024 * 1024  # 1 gigabyte
        self.BROADCAST_ADDRS = ["<broadcast>", "0.0.0.0"]
        self.PREFER_IP_VERSION = 0  # 4, 6 or 0 (0=let OS choose according to RFC 3484)
        self.SERIALIZER = "serpent"
        self.SERPENT_BYTES_REPR = False
        self.LOGWIRE = False
        self.ITER_STREAMING = True
        self.ITER_STREAM_LIFETIME = 0.0
        self.ITER_STREAM_LINGER = 30.0
        self.LOGFILE = _pyro_logfile
        self.LOGLEVEL = _pyro_loglevel
        self.SSL = False
        self.SSL_SERVERCERT = ""
        self.SSL_SERVERKEY = ""
        self.SSL_SERVERKEYPASSWD = ""
        self.SSL_REQUIRECLIENTCERT = False
        self.SSL_CLIENTCERT = ""
        self.SSL_CLIENTKEY = ""
        self.SSL_CLIENTKEYPASSWD = ""
        self.SSL_CACERTS = ""
        if use_environment:
            # environment variables overwrite config items
            prefix = "PYRO_"
            for item, envvalue in (e for e in os.environ.items() if e[0].startswith(prefix)):
                item = item[len(prefix):]
                if item not in self.__slots__:
                    raise ValueError("invalid Pyro environment config variable: %s%s" % (prefix, item))
                value = getattr(self, item)
                valuetype = type(value)
                if valuetype is set:
                    envvalue = {v.strip() for v in envvalue.split(",")}
                elif valuetype is list:
                    envvalue = [v.strip() for v in envvalue.split(",")]
                elif valuetype is bool:
                    envvalue = envvalue.lower()
                    if envvalue in ("0", "off", "no", "false"):
                        envvalue = False
                    elif envvalue in ("1", "yes", "on", "true"):
                        envvalue = True
                    else:
                        raise ValueError("invalid boolean value: %s%s=%s" % (prefix, item, envvalue))
                else:
                    try:
                        envvalue = valuetype(envvalue)
                    except ValueError:
                        raise ValueError("invalid Pyro environment config value: %s%s=%s" % (prefix, item, envvalue)) from None
                setattr(self, item, envvalue)

    def copy(self):
        """returns a copy of this config"""
        other = object.__new__(Configuration)
        for item in self.__slots__:
            setattr(other, item, getattr(self, item))
        return other

    def as_dict(self):
        """returns this config as a regular dictionary"""
        return {item: getattr(self, item) for item in self.__slots__}

    def dump(self):
        """Easy config diagnostics"""
        from .protocol import PROTOCOL_VERSION
        result = ["Pyro version: %s" % __version__,
                  "Loaded from: %s" % os.path.dirname(__file__),
                  "Python version: %s %s (%s, %s)" % (platform.python_implementation(),
                                                      platform.python_version(), platform.system(), os.name),
                  "Protocol version: %d" % PROTOCOL_VERSION,
                  "Currently active global configuration settings:"]
        for item, value in sorted(self.as_dict().items()):
            result.append("{:s} = {:s}".format(item, str(value)))
        return "\n".join(result)


global_config = Configuration()


def dump():
    print(global_config.dump())


if __name__ == "__main__":
    dump()



from Pyro5.api import Proxy
import Pyro5.errors


print("First start the built-in test echo server with something like:")
print("$ python -m Pyro5.utils.echoserver")
print("Enter the server's uri that was printed:")
uri = input().strip()
echoserver = Proxy(uri)

response = echoserver.echo("hello")
print("\ngot back from the server: %s" % response)
response = echoserver.echo([1, 2, 3, 4])
print("got back from the server: %s" % response)

for element in echoserver.generator():
    print("got element from remote iterator:", element)

try:
    echoserver.error()
except Exception:
    print("\ncaught an exception (expected), traceback:")
    print("".join(Pyro5.errors.get_pyro_traceback()))

print("\nshutting down the test echo server. (restart it if you want to run this again)")
echoserver.shutdown()



import pendulum
import datetime
from Pyro5.api import Proxy


uri = input("What is the server uri? ").strip()
fmt = '%Y-%m-%d %H:%M:%S %Z%z'
print("local time without timezone: ", datetime.datetime.now().strftime(fmt))


with Proxy(uri) as serv:
    print("\n1. no timezone")
    datestr = serv.echo(datetime.datetime.now())
    print("Got from server:", repr(datestr))
    dt = pendulum.parse(datestr)
    print("   parsed:", repr(dt))

    print("\n2. PyTz timezones")
    datestr = serv.pytz()
    print("Got from server:", repr(datestr))
    dt = pendulum.parse(datestr)
    print("   parsed:", repr(dt))

    print("\n3. DateUtil timezones")
    datestr = serv.dateutil()
    print("Got from server:", repr(datestr))
    dt = pendulum.parse(datestr)
    print("   parsed:", repr(dt))

    print("\n4. Pendulum timezones")
    datestr = serv.pendulum()
    print("Got from server:", repr(datestr))
    dt = pendulum.parse(datestr)
    print("   parsed:", repr(dt))
    print()



from Pyro5.api import Proxy


with Proxy("PYRONAME:example.eventloop.server") as proxy:
    print("5*11=%d" % proxy.multiply(5, 11))
    print("'x'*10=%s" % proxy.multiply('x', 10))

    input("press enter to do a loop of some more calls:")
    for i in range(1, 20):
        print("2*i=%d" % proxy.multiply(2, i))
        print("'@'*i=%s" % proxy.multiply('@', i))



from Pyro5.api import behavior, expose, locate_ns, serve, config
from diffiehellman import DiffieHellman


config.SERVERTYPE = "multiplex"

ns = locate_ns()


@behavior(instance_mode="session")
class KeyExchange(object):
    def __init__(self):
        print("New KeyExchange, initializing Diffie-Hellman")
        self.dh = DiffieHellman(group=14)

    @expose
    def exchange_key(self, other_public_key):
        print("received a public key, calculating shared secret...")
        self.dh.make_shared_secret_and_key(other_public_key)
        print("shared secret key = ", self.dh.key.hex())
        return self.dh.public_key


serve({
    KeyExchange: "example.dh.keyexchange"
}, use_ns=True)



"""
This is the subscriber for the 'many topics' messages example.
For code with more explanations, see the regular 'weather' message example code.
"""
import os
import time
import threading
from operator import itemgetter
from Pyro5.api import expose, Daemon
from messagebus.messagebus import Subscriber


@expose
class Subber(Subscriber):
    def init_counters(self, topics):
        self.message_counter = {}
        self.last_message = {}
        for t in topics:
            self.message_counter[t] = 0
            self.last_message[t] = None

    def consume_message(self, topic, message):
        self.message_counter[topic] += 1
        self.last_message[topic] = message


def clear_screen():
    os.system(['clear', 'cls'][os.name == 'nt'])


subber = Subber()
d = Daemon()
d.register(subber)
daemon_thread = threading.Thread(target=d.requestLoop)
daemon_thread.daemon = True
daemon_thread.start()

# mass subscribe to all available topics
topics = list(sorted(subber.bus.topics()))
subber.init_counters(topics)
for t in topics:
    subber.bus.subscribe(t, subber)

# show a table of the active topics on the bus
while True:
    clear_screen()
    print(time.ctime(), "-- active topics on the messagebus:")
    print("{:20} : {:5}     {}         {}".format("topic", "count", "last_recv", "last message data"))
    for topic, count in sorted(subber.message_counter.items(), key=itemgetter(1), reverse=True):
        msg = subber.last_message[topic]
        if msg:
            print("{:20} : {:5d}  -  {}   {!r:.20}".format(topic, count, msg.created.time(), msg.data))
        else:
            print("{:20} : {:5d}".format(topic, count))
    print("(restart me to refresh the list of topics)")
    time.sleep(1)



"""
Tests for a running Pyro server, without timeouts.

Pyro - Python Remote Objects.  Copyright by Irmen de Jong (irmen@razorvine.net).
"""

import time
import threading
import serpent
import pytest
import Pyro5.core
import Pyro5.client
import Pyro5.server
import Pyro5.errors
import Pyro5.serializers
import Pyro5.protocol
import Pyro5.callcontext
import Pyro5.socketutil
from Pyro5 import config
from support import *


@Pyro5.server.expose
class ServerTestObject(object):
    something = 99
    dict_attr = {}

    def __init__(self):
        self._dictionary = {"number": 42}
        self.dict_attr = {"number2": 43}
        self._value = 12345
        self.items = list("qwerty")

    def getDict(self):
        return self._dictionary

    def getDictAttr(self):
        return self.dict_attr

    def multiply(self, x, y):
        return x * y

    def divide(self, x, y):
        return x // y

    def ping(self):
        pass

    def echo(self, obj):
        return obj

    def blob(self, blob):
        return blob.info, blob.deserialized()

    @Pyro5.server.oneway
    def oneway_delay(self, delay):
        time.sleep(delay)

    def delay(self, delay):
        time.sleep(delay)
        return "slept %d seconds" % delay

    def delayAndId(self, delay, id):
        time.sleep(delay)
        return "slept for " + str(id)

    def testargs(self, x, *args, **kwargs):
        return [x, list(args), kwargs]  # don't return tuples, this enables us to test json serialization as well.

    def nonserializableException(self):
        raise NonserializableError(("xantippe", lambda x: 0))

    @Pyro5.server.oneway
    def oneway_multiply(self, x, y):
        return x * y

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, newvalue):
        self._value = newvalue

    @property
    def dictionary(self):
        return self._dictionary

    def iterator(self):
        return iter(["one", "two", "three"])

    def generator(self):
        yield "one"
        yield "two"
        yield "three"
        yield "four"
        yield "five"

    def response_annotation(self):
        # part of the annotations tests
        if "XYZZ" not in Pyro5.callcontext.current_context.annotations:
            raise ValueError("XYZZ should be present in annotations in the daemon")
        if Pyro5.callcontext.current_context.annotations["XYZZ"] != b"data from proxy via new api":
            raise ValueError("XYZZ annotation has wrong data")
        Pyro5.callcontext.current_context.response_annotations["ANN2"] = b"daemon annotation via new api"
        return {"annotations_in_daemon": Pyro5.callcontext.current_context.annotations}

    def new_test_object(self):
        return ServerTestObject()

    def __len__(self):
        return len(self.items)

    def __iter__(self):
        return iter(self.items)

    def __getitem__(self, item):
        return self.items[item]


class NotEverythingExposedClass(object):
    def __init__(self, name):
        self.name = name

    @Pyro5.server.expose
    def getName(self):
        return self.name

    def unexposed(self):
        return "you should not see this"


class DaemonLoopThread(threading.Thread):
    def __init__(self, pyrodaemon):
        super().__init__()
        self.daemon = True
        self.pyrodaemon = pyrodaemon
        self.running = threading.Event()
        self.running.clear()

    def run(self):
        self.running.set()
        try:
            self.pyrodaemon.requestLoop()
        except Pyro5.errors.CommunicationError:
            pass  # ignore pyro communication errors


class DaemonWithSabotagedHandshake(Pyro5.server.Daemon):
    def _handshake(self, conn, denied_reason=None):
        # receive the client's handshake data
        msg = Pyro5.protocol.recv_stub(conn, [Pyro5.protocol.MSG_CONNECT])
        # return a CONNECTFAIL always
        serializer = Pyro5.serializers.serializers_by_id[msg.serializer_id]
        data = serializer.dumps("rigged connection failure")
        msg = Pyro5.protocol.SendingMessage(Pyro5.protocol.MSG_CONNECTFAIL, 0, 1, serializer.serializer_id, data)
        conn.send(msg.data)
        return False


class TestServerBrokenHandshake:
    def setup_method(self):
        config.LOGWIRE = True
        self.daemon = DaemonWithSabotagedHandshake(port=0)
        obj = ServerTestObject()
        uri = self.daemon.register(obj, "something")
        self.objectUri = uri
        self.daemonthread = DaemonLoopThread(self.daemon)
        self.daemonthread.start()
        self.daemonthread.running.wait()
        time.sleep(0.05)

    def teardown_method(self):
        time.sleep(0.05)
        self.daemon.shutdown()
        self.daemonthread.join()

    def testDaemonConnectFail(self):
        # check what happens when the daemon responds with a failed connection msg
        with Pyro5.client.Proxy(self.objectUri) as p:
            with pytest.raises(Pyro5.errors.CommunicationError) as x:
                p.ping()
            message = str(x.value)
            assert "rejected:" in message
            assert "rigged connection failure" in message


class TestServerOnce:
    """tests that are fine to run with just a single server type"""

    def setup_method(self):
        config.SERIALIZER = "serpent"
        config.LOGWIRE = True
        Pyro5.serializers.SerializerBase.register_class_to_dict(ServerTestObject, lambda x: {})
        Pyro5.serializers.SerializerBase.register_dict_to_class("test_server.ServerTestObject", lambda cn, d: ServerTestObject())
        self.daemon = Pyro5.server.Daemon(port=0)
        obj = ServerTestObject()
        uri = self.daemon.register(obj, "something")
        self.objectUri = uri
        obj2 = NotEverythingExposedClass("hello")
        self.daemon.register(obj2, "unexposed")
        self.daemonthread = DaemonLoopThread(self.daemon)
        self.daemonthread.start()
        self.daemonthread.running.wait()
        time.sleep(0.05)

    def teardown_method(self):
        time.sleep(0.05)
        Pyro5.serializers.SerializerBase.unregister_class_to_dict(ServerTestObject)
        Pyro5.serializers.SerializerBase.unregister_dict_to_class("test_server.ServerTestObject")
        if self.daemon is not None:
            self.daemon.shutdown()
            self.daemonthread.join()

    def testPingMessage(self):
        with Pyro5.client.Proxy(self.objectUri) as p:
            p._pyroBind()
            conn = p._pyroConnection
            msg = Pyro5.protocol.SendingMessage(Pyro5.protocol.MSG_PING, 42, 999, 0, b"something")
            conn.send(msg.data)
            msg = Pyro5.protocol.recv_stub(conn, [Pyro5.protocol.MSG_PING])
            assert msg.type == Pyro5.protocol.MSG_PING
            assert msg.seq == 999
            assert msg.data == b"pong"
            Pyro5.protocol.SendingMessage.ping(p._pyroConnection)  # the convenience method that does the above

    def testSequence(self):
        with Pyro5.client.Proxy(self.objectUri) as p:
            p.echo(1)
            p.echo(2)
            p.echo(3)
            assert p._pyroSeq, "should have 3 method calls" == 3
            p._pyroSeq = 999   # hacking the seq nr won't have any effect because it is the reply from the server that is checked
            assert p.echo(42) == 42

    def testMetaOnAttrs(self):
        with Pyro5.client.Proxy(self.objectUri) as p:
            assert p.multiply(5, 11) == 55
            # property
            x = p.getDict()
            assert x == {"number": 42}
            p.dictionary.update({"more": 666})  # should not fail because metadata is enabled and the dictionary property is retrieved as local copy
            x = p.getDict()
            assert x == {"number": 42}  # not updated remotely because we had a local copy
        with Pyro5.client.Proxy(self.objectUri) as p:
            with pytest.raises(AttributeError):
                # attribute should fail (meta only works for exposed properties)
                p.dict_attr.update({"more": 666})

    def testSomeArgumentTypes(self):
        with Pyro5.client.Proxy(self.objectUri) as p:
            assert p.testargs(1) == [1, [], {}]
            assert p.testargs(1, 2, 3, a=4) == [1, [2, 3], {'a': 4}]
            assert p.testargs(1, **{'a': 2}) == [1, [], {'a': 2}]

    def testUnicodeKwargs(self):
        with Pyro5.client.Proxy(self.objectUri) as p:
            assert p.testargs(1, **{chr(65): 2}) == [1, [], {chr(65): 2}]
            result = p.testargs(chr(0x20ac), **{chr(0x20ac): 2})
            assert chr(0x20ac) == result[0]
            key = list(result[2].keys())[0]
            assert chr(0x20ac) == key

    def testNormalProxy(self):
        with Pyro5.client.Proxy(self.objectUri) as p:
            assert p.multiply(7, 6) == 42

    def testExceptions(self):
        with Pyro5.client.Proxy(self.objectUri) as p:
            with pytest.raises(ZeroDivisionError):
                p.divide(1, 0)
            with pytest.raises(TypeError):
                p.multiply("a", "b")

    def testProxyMetadata(self):
        with Pyro5.client.Proxy(self.objectUri) as p:
            # unconnected proxies have empty metadata
            assert p._pyroAttrs == set()
            assert p._pyroMethods == set()
            assert p._pyroOneway == set()
            # connecting it should obtain metadata
            p._pyroBind()
            assert p._pyroAttrs == {'value', 'dictionary'}
            assert p._pyroMethods == {'echo', 'getDict', 'divide', 'nonserializableException', 'ping', 'oneway_delay', 'delayAndId', 'delay', 'testargs',
                              'multiply', 'oneway_multiply', 'getDictAttr', 'iterator', 'generator', 'response_annotation', 'blob', 'new_test_object',
                              '__iter__', '__len__', '__getitem__'}
            assert p._pyroOneway == {'oneway_multiply', 'oneway_delay'}
            p._pyroAttrs = None
            p._pyroGetMetadata()
            assert p._pyroAttrs == {'value', 'dictionary'}
            p._pyroAttrs = None
            p._pyroGetMetadata(self.objectUri.object)
            assert p._pyroAttrs == {'value', 'dictionary'}
            p._pyroAttrs = None
            p._pyroGetMetadata(known_metadata={"attrs": set(), "oneway": set(), "methods": {"ping"}})
            assert p._pyroAttrs == set()

    def testProxyAttrsMetadataOn(self):
        # read attributes
        with Pyro5.client.Proxy(self.objectUri) as p:
            # unconnected proxy still has empty metadata.
            # but, as soon as an attribute is used, the metadata is obtained
            a = p.value
            assert a == 12345
            a = p.multiply
            assert isinstance(a, Pyro5.client._RemoteMethod)  # multiply is still a regular method
            with pytest.raises(AttributeError):
                _ = p.non_existing_attribute
        # set attributes, should also trigger getting metadata
        with Pyro5.client.Proxy(self.objectUri) as p:
            p.value = 42
            assert p.value == 42
            assert "value" in p._pyroAttrs

    def testProxyAnnotations(self):
        with Pyro5.client.Proxy(self.objectUri) as p:
            Pyro5.callcontext.current_context.annotations = {"XYZZ": b"invalid test data"}
            with pytest.raises(ValueError):
                p.response_annotation()
            Pyro5.callcontext.current_context.annotations = {"XYZZ": b"data from proxy via new api"}
            response = p.response_annotation()
            assert Pyro5.callcontext.current_context.response_annotations["ANN2"] == b"daemon annotation via new api"
            # check that the daemon received both the old and the new annotation api data:
            daemon_annotations = response["annotations_in_daemon"]
            assert serpent.tobytes(daemon_annotations["XYZZ"]) == b"data from proxy via new api"

    def testExposedRequired(self):
        with self.daemon.proxyFor("unexposed") as p:
            assert p._pyroMethods == {"getName"}
            assert p.getName() == "hello"
            with pytest.raises(AttributeError) as e:
                p.unexposed()
            expected_msg = "remote object '%s' has no exposed attribute or method 'unexposed'" % p._pyroUri
            assert str(e.value) == expected_msg
            with pytest.raises(AttributeError) as e:
                p.unexposed_set = 999
            expected_msg = "remote object '%s' has no exposed attribute 'unexposed_set'" % p._pyroUri
            assert str(e.value) == expected_msg

    def testProperties(self):
        with Pyro5.client.Proxy(self.objectUri) as p:
            _ = p.value
            # metadata should be loaded now
            assert p._pyroAttrs == {"value", "dictionary"}
            with pytest.raises(AttributeError):
                _ = p.something
            with pytest.raises(AttributeError):
                _ = p._dictionary
            with pytest.raises(AttributeError):
                _ = p._value
            assert p.value == 12345
            assert p.dictionary == {"number": 42}

    def testHasAttr(self):
        with Pyro5.client.Proxy(self.objectUri) as p:
            # with metadata on, hasattr actually gives proper results
            assert hasattr(p, "multiply")
            assert hasattr(p, "oneway_multiply")
            assert hasattr(p, "value")
            assert not hasattr(p, "_value")
            assert not hasattr(p, "_dictionary")
            assert not hasattr(p, "non_existing_attribute")

    def testProxyMetadataKnown(self):
        with Pyro5.client.Proxy(self.objectUri) as p:
            # unconnected proxies have empty metadata
            assert p._pyroAttrs == set()
            assert p._pyroMethods == set()
            assert p._pyroOneway == set()
            # set some metadata manually, they should be overwritten at connection time
            p._pyroMethods = set("abc")
            p._pyroAttrs = set("xyz")
            p._pyroBind()
            assert p._pyroAttrs != set("xyz")
            assert p._pyroMethods != set("abc")
            assert p._pyroOneway != set()

    def testNonserializableException_other(self):
        with Pyro5.client.Proxy(self.objectUri) as p:
            with pytest.raises(Pyro5.errors.PyroError) as x:
                p.nonserializableException()
            tblines = "\n".join(Pyro5.errors.get_pyro_traceback(x.type, x.value, x.tb))
            assert "unsupported serialized class" in tblines

    def testBatchProxy(self):
        with Pyro5.client.Proxy(self.objectUri) as p:
            batch = Pyro5.client.BatchProxy(p)
            assert batch.multiply(7, 6) is None
            assert batch.divide(999, 3) is None
            assert batch.ping() is None
            assert batch.divide(999, 0) is None    # force an error here
            assert batch.multiply(3, 4) is None    # this call should not be performed anymore
            results = batch()
            assert next(results) == 42
            assert next(results) == 333
            assert next(results) is None
            with pytest.raises(ZeroDivisionError):
                next(results)    # 999//0 should raise this error
            with pytest.raises(StopIteration):
                next(results)    # no more results should be available after the error

    def testBatchOneway(self):
        with Pyro5.client.Proxy(self.objectUri) as p:
            batch = Pyro5.client.BatchProxy(p)
            assert batch.multiply(7, 6) is None
            assert batch.delay(1) is None    # a delay shouldn't matter with oneway
            assert batch.multiply(3, 4) is None
            begin = time.time()
            results = batch(oneway=True)
            duration = time.time() - begin
            assert duration < 0.1, "oneway batch with delay should return almost immediately"
            assert results is None

    def testPyroTracebackNormal(self):
        with Pyro5.client.Proxy(self.objectUri) as p:
            with pytest.raises(ZeroDivisionError) as x:
                p.divide(999, 0)  # force error here
            # going to check if the magic pyro traceback attribute is available for batch methods too
            tb = "".join(Pyro5.errors.get_pyro_traceback(x.type, x.value, x.tb))
            assert "Remote traceback:" in tb  # validate if remote tb is present
            assert "ZeroDivisionError" in tb  # the error
            assert "return x // y" in tb  # the statement

    def testPyroTracebackBatch(self):
        with Pyro5.client.Proxy(self.objectUri) as p:
            batch = Pyro5.client.BatchProxy(p)
            assert batch.divide(999, 0) is None    # force an exception here
            results = batch()
            with pytest.raises(ZeroDivisionError) as x:
                next(results)
            # going to check if the magic pyro traceback attribute is available for batch methods too
            tb = "".join(Pyro5.errors.get_pyro_traceback(x.type, x.value, x.tb))
            assert "Remote traceback:" in tb  # validate if remote tb is present
            assert "ZeroDivisionError" in tb  # the error
            assert "return x // y" in tb  # the statement
            with pytest.raises(StopIteration):
                next(results)  # no more results should be available after the error

    def testAutoProxy(self):
        obj = ServerTestObject()
        with Pyro5.client.Proxy(self.objectUri) as p:
            result = p.echo(obj)
            assert isinstance(result, ServerTestObject), "non-pyro object must be returned as normal class"
            self.daemon.register(obj)
            result = p.echo(obj)
            assert isinstance(result, Pyro5.client.Proxy), "serialized pyro object must be a proxy"
            self.daemon.register(ServerTestObject)
            new_result = result.new_test_object()
            assert isinstance(new_result, Pyro5.client.Proxy), "serialized pyro object must be a proxy"
            self.daemon.unregister(ServerTestObject)
            self.daemon.unregister(obj)
            result = p.echo(obj)
            assert isinstance(result, ServerTestObject), "unregistered pyro object must be normal class again"

    def testRegisterWeak(self):
        obj=ServerTestObject()
        uri=self.daemon.register(obj,weak=True)
        with Pyro5.client.Proxy(uri) as p:
            result = p.getDict()
            assert isinstance(result, dict), "getDict() is proxied normally"
            del obj # weak registration should not prevent the obj from being garbage-collected
            with pytest.raises(Pyro5.errors.DaemonError):
                result = p.getDict()

    def testConnectOnce(self):
        with Pyro5.client.Proxy(self.objectUri) as proxy:
            assert proxy._pyroBind(), "first bind should always connect"
            assert not proxy._pyroBind(), "second bind should not connect again"

    def testMaxMsgSize(self):
        with Pyro5.client.Proxy(self.objectUri) as p:
            bigobject = [42] * 1000
            result = p.echo(bigobject)
            assert bigobject == result
            try:
                config.MAX_MESSAGE_SIZE = 999
                with pytest.raises(Pyro5.errors.ProtocolError):
                    _ = p.echo(bigobject)       # message too large
            finally:
                config.MAX_MESSAGE_SIZE = 1024* 1024* 1024

    def testIterator(self):
        with Pyro5.client.Proxy(self.objectUri) as p:
            iterator = p.iterator()
            assert isinstance(iterator, Pyro5.client._StreamResultIterator)
            assert next(iterator) == "one"
            assert next(iterator) == "two"
            assert next(iterator) == "three"
            with pytest.raises(StopIteration):
                next(iterator)
            iterator.close()

    def testLenAndIterAndIndexing(self):
        with Pyro5.client.Proxy(self.objectUri) as p:
            assert len(p) == 6
            values = list(iter(p))
            assert values == ['q', 'w', 'e', 'r', 't', 'y']
            assert p[0] == 'q'
            assert p[1] == 'w'
            assert p[2] == 'e'
            assert p[3] == 'r'
            assert p[4] == 't'
            assert p[5] == 'y'
            with pytest.raises(IndexError):
                _ = p[6]

    def testGenerator(self):
        with Pyro5.client.Proxy(self.objectUri) as p:
            generator = p.generator()
            assert isinstance(generator, Pyro5.client._StreamResultIterator)
            assert next(generator) == "one"
            assert next(generator) == "two"
            assert next(generator) == "three"
            assert next(generator) == "four"
            assert next(generator) == "five"
            with pytest.raises(StopIteration):
                next(generator)
            with pytest.raises(StopIteration):
                next(generator)
            generator.close()
            generator = p.generator()
            _ = [v for v in generator]
            with pytest.raises(StopIteration):
                next(generator)
            generator.close()

    def testCleanup(self):
        p1 = Pyro5.client.Proxy(self.objectUri)
        p2 = Pyro5.client.Proxy(self.objectUri)
        p3 = Pyro5.client.Proxy(self.objectUri)
        p1.echo(42)
        p2.echo(42)
        p3.echo(42)
        # we have several active connections still up, see if we can cleanly shutdown the daemon
        # (it should interrupt the worker's socket connections)
        time.sleep(0.1)
        self.daemon.shutdown()
        self.daemon = None
        p1._pyroRelease()
        p2._pyroRelease()
        p3._pyroRelease()

    def testSerializedBlob(self):
        sb = Pyro5.client.SerializedBlob("blobname", [1, 2, 3])
        assert sb.info == "blobname"
        assert sb.deserialized() == [1, 2, 3]

    def testSerializedBlobMessage(self):
        serializer = Pyro5.serializers.serializers["serpent"]
        data = serializer.dumpsCall("object", "method", ([1, 2, 3],), {"kwarg": 42})
        msg = Pyro5.protocol.SendingMessage(Pyro5.protocol.MSG_INVOKE, 0, 42, serializer.serializer_id, data)
        sb = Pyro5.client.SerializedBlob("blobname", msg, is_blob=True)
        assert sb.info == "blobname"
        assert sb.deserialized() == ([1, 2, 3], )

    def testProxySerializedBlobArg(self):
        with Pyro5.client.Proxy(self.objectUri) as p:
            blobinfo, blobdata = p.blob(Pyro5.client.SerializedBlob("blobname", [1, 2, 3]))
            assert blobinfo == "blobname"
            assert blobdata == [1, 2, 3]

    def testResourceFreeing(self):
        rsvc = ResourceService()
        uri = self.daemon.register(rsvc)
        with Pyro5.client.Proxy(uri) as p:
            p.allocate("r1")
            p.allocate("r2")
            resources = {r.name: r for r in rsvc.resources}
            p.free("r1")
            rsc = p.list()
            assert rsc == ["r2"]
            assert resources["r1"].close_called
            assert not resources["r2"].close_called
        time.sleep(0.02)
        assert resources["r1"].close_called
        assert resources["r2"].close_called
        with Pyro5.client.Proxy(uri) as p:
            rsc = p.list()
            assert rsc == [], "r2 must now be freed due to connection loss earlier"



class TestServerThreadNoTimeout:
    SERVERTYPE = "thread"
    COMMTIMEOUT = None

    def setup_method(self):
        config.SERIALIZER = "serpent"
        config.LOGWIRE = True
        config.POLLTIMEOUT = 0.1
        config.SERVERTYPE = self.SERVERTYPE
        config.COMMTIMEOUT = self.COMMTIMEOUT
        self.daemon = Pyro5.server.Daemon(port=0)
        obj = ServerTestObject()
        uri = self.daemon.register(obj, "something")
        self.objectUri = uri
        self.daemonthread = DaemonLoopThread(self.daemon)
        self.daemonthread.start()
        self.daemonthread.running.wait()
        time.sleep(0.05)

    def teardown_method(self):
        time.sleep(0.05)
        self.daemon.shutdown()
        self.daemonthread.join()
        config.SERVERTYPE = "thread"
        config.COMMTIMEOUT = None

    def testConnectionStuff(self):
        p1 = Pyro5.client.Proxy(self.objectUri)
        p2 = Pyro5.client.Proxy(self.objectUri)
        assert not p1._pyroConnection
        assert not p2._pyroConnection
        p1.ping()
        p2.ping()
        _ = p1.multiply(11, 5)
        _ = p2.multiply(11, 5)
        assert p1._pyroConnection
        assert p2._pyroConnection
        p1._pyroRelease()
        p1._pyroRelease()
        p2._pyroRelease()
        p2._pyroRelease()
        assert not p1._pyroConnection
        assert not p2._pyroConnection
        p1._pyroBind()
        _ = p1.multiply(11, 5)
        _ = p2.multiply(11, 5)
        assert p1._pyroConnection
        assert p2._pyroConnection
        assert p1._pyroUri.protocol == "PYRO"
        assert p2._pyroUri.protocol == "PYRO"
        p1._pyroRelease()
        p2._pyroRelease()

    def testReconnectAndCompression(self):
        # try reconnects
        with Pyro5.client.Proxy(self.objectUri) as p:
            assert not p._pyroConnection
            p._pyroReconnect(tries=100)
            assert p._pyroConnection
        assert not p._pyroConnection
        # test compression:
        try:
            with Pyro5.client.Proxy(self.objectUri) as p:
                config.COMPRESSION = True
                assert p.multiply(5, 11) == 55
                assert p.multiply("*" * 500, 2) == "*" * 1000
        finally:
            config.COMPRESSION = False

    def testOnewayMetaOn(self):
        with Pyro5.client.Proxy(self.objectUri) as p:
            assert p._pyroOneway == set()  # when not bound, no meta info exchange has been done
            p._pyroBind()
            assert "oneway_multiply" in p._pyroOneway  # after binding, meta info has been processed
            assert p.multiply(5, 11) == 55  # not tagged as @Pyro5.server.oneway
            assert p.oneway_multiply(5, 11) is None   # tagged as @Pyro5.server.oneway
            p._pyroOneway = set()
            assert p.multiply(5, 11) == 55
            assert p.oneway_multiply(5, 11) == 55
            # check nonexisting method behavoir for oneway methods
            with pytest.raises(AttributeError):
                p.nonexisting_method()
            p._pyroOneway.add("nonexisting_method")
            # now it should still fail because of metadata telling Pyro what methods actually exist
            with pytest.raises(AttributeError):
                p.nonexisting_method()

    def testOnewayWithProxySubclass(self):
        class ProxyWithOneway(Pyro5.client.Proxy):
            def __init__(self, arg):
                super(ProxyWithOneway, self).__init__(arg)
                self._pyroOneway = {"oneway_multiply", "multiply"}

        with ProxyWithOneway(self.objectUri) as p:
            assert p.oneway_multiply(5, 11) is None
            assert p.multiply(5, 11) == 55
            p._pyroOneway = set()
            assert p.oneway_multiply(5, 11) == 55
            assert p.multiply(5, 11) == 55

    def testOnewayDelayed(self):
        with Pyro5.client.Proxy(self.objectUri) as p:
            p.ping()
            now = time.time()
            p.oneway_delay(1)  # oneway so we should continue right away
            time.sleep(0.01)
            assert time.time() - now < 0.2, "delay should be running as oneway"
            now = time.time()
            assert p.multiply(5, 11), "expected a normal result from a non-oneway call" == 55
            assert time.time() - now < 0.2, "delay should be running in its own thread"

    def testSerializeConnected(self):
        # online serialization tests
        ser = Pyro5.serializers.serializers[config.SERIALIZER]
        proxy = Pyro5.client.Proxy(self.objectUri)
        proxy._pyroBind()
        assert proxy._pyroConnection
        p = ser.dumps(proxy)
        proxy2 = ser.loads(p)
        assert proxy2._pyroConnection is None
        assert proxy._pyroConnection
        assert proxy._pyroUri == proxy2._pyroUri
        proxy2._pyroBind()
        assert proxy2._pyroConnection
        assert proxy2._pyroConnection is not proxy._pyroConnection
        proxy._pyroRelease()
        proxy2._pyroRelease()
        assert proxy._pyroConnection is None
        assert proxy2._pyroConnection is None
        proxy.ping()
        proxy2.ping()
        # try copying a connected proxy
        import copy
        proxy3 = copy.copy(proxy)
        assert proxy3._pyroConnection is None
        assert proxy._pyroConnection
        assert proxy._pyroUri == proxy3._pyroUri
        assert proxy3._pyroUri is not proxy._pyroUri
        proxy._pyroRelease()
        proxy2._pyroRelease()
        proxy3._pyroRelease()

    def testException(self):
        with Pyro5.client.Proxy(self.objectUri) as p:
            with pytest.raises(ZeroDivisionError) as x:
                p.divide(1, 0)
            pyrotb = "".join(Pyro5.errors.get_pyro_traceback(x.type, x.value, x.tb))
            assert "Remote traceback" in pyrotb
            assert "ZeroDivisionError" in pyrotb

    def testTimeoutCall(self):
        config.COMMTIMEOUT = None
        with Pyro5.client.Proxy(self.objectUri) as p:
            p.ping()
            start = time.time()
            p.delay(0.5)
            duration = time.time() - start
            assert 0.4 < duration < 0.6
            p._pyroTimeout = 0.1
            start = time.time()
            with pytest.raises(Pyro5.errors.TimeoutError):
                p.delay(1)
            duration = time.time() - start
            assert duration < 0.3

    def testTimeoutConnect(self):
        # set up a unresponsive daemon
        with Pyro5.server.Daemon(port=0) as d:
            time.sleep(0.5)
            obj = ServerTestObject()
            uri = d.register(obj)
            # we're not going to start the daemon's event loop
            p = Pyro5.client.Proxy(uri)
            p._pyroTimeout = 0.2
            start = time.time()
            with pytest.raises(Pyro5.errors.TimeoutError) as e:
                p.ping()
            assert str(e.value) == "receiving: timeout"

    # XXX todo: add test about proxy thread ownership transfer

    def testServerConnections(self):
        # check if the server allows to grow the number of connections
        proxies = [Pyro5.client.Proxy(self.objectUri) for _ in range(10)]
        try:
            for p in proxies:
                p._pyroTimeout = 0.5
                p._pyroBind()
            for p in proxies:
                p.ping()
        finally:
            for p in proxies:
                p._pyroRelease()

    def testGeneratorProxyClose(self):
        p = Pyro5.client.Proxy(self.objectUri)
        generator = p.generator()
        p._pyroRelease()
        with pytest.raises(Pyro5.errors.ConnectionClosedError):
            next(generator)

    def testGeneratorLinger(self):
        orig_linger = config.ITER_STREAM_LINGER
        orig_commt = config.COMMTIMEOUT
        orig_pollt = config.POLLTIMEOUT
        try:
            config.ITER_STREAM_LINGER = 0.5
            config.COMMTIMEOUT = 0.2
            config.POLLTIMEOUT = 0.2
            p = Pyro5.client.Proxy(self.objectUri)
            generator = p.generator()
            assert next(generator) == "one"
            p._pyroRelease()
            with pytest.raises(Pyro5.errors.ConnectionClosedError):
                next(generator)
            p._pyroReconnect()
            assert next(generator), "generator should resume after reconnect" == "two"
            # check that after the linger time passes, the generator *is* gone
            p._pyroRelease()
            time.sleep(2)
            p._pyroReconnect()
            with pytest.raises(Pyro5.errors.PyroError):  # should not be resumable anymore
                next(generator)
        finally:
            config.ITER_STREAM_LINGER = orig_linger
            config.COMMTIMEOUT = orig_commt
            config.POLLTIMEOUT = orig_pollt

    def testGeneratorNoLinger(self):
        orig_linger = config.ITER_STREAM_LINGER
        try:
            p = Pyro5.client.Proxy(self.objectUri)
            config.ITER_STREAM_LINGER = 0  # disable linger
            generator = p.generator()
            assert next(generator) == "one"
            p._pyroRelease()
            time.sleep(0.2)
            with pytest.raises(Pyro5.errors.ConnectionClosedError):
                next(generator)
            p._pyroReconnect()
            with pytest.raises(Pyro5.errors.PyroError):  # should not be resumable after reconnect
                next(generator)
            generator.close()
        finally:
            config.ITER_STREAM_LINGER = orig_linger


class TestServerMultiplexNoTimeout(TestServerThreadNoTimeout):
    SERVERTYPE = "multiplex"
    COMMTIMEOUT = None

    def testException(self):
        pass


class TestMetaAndExpose:
    def testBasic(self):
        o = MyThingFullExposed("irmen")
        m1 = Pyro5.server._get_exposed_members(o)
        m2 = Pyro5.server._get_exposed_members(MyThingFullExposed)
        assert m1 == m2
        keys = m1.keys()
        assert len(keys) == 3
        assert "methods" in keys
        assert "attrs" in keys
        assert "oneway" in keys

    def testGetExposedCacheWorks(self):
        class Thingy(object):
            def method1(self):
                pass
            @property
            def prop(self):
                return 1
            def notexposed(self):
                pass
        m1 = Pyro5.server._get_exposed_members(Thingy, only_exposed=False)
        def new_method(self, arg):
            return arg
        Thingy.new_method = new_method
        m2 = Pyro5.server._get_exposed_members(Thingy, only_exposed=False)
        assert m2, "should still be equal because result from cache" == m1

    def testPrivateNotExposed(self):
        o = MyThingFullExposed("irmen")
        m = Pyro5.server._get_exposed_members(o)
        assert m["methods"] == {"classmethod", "staticmethod", "method", "__dunder__", "oneway", "exposed"}
        assert m["attrs"] == {"prop1", "readonly_prop1", "prop2"}
        assert m["oneway"] == {"oneway"}
        o = MyThingPartlyExposed("irmen")
        m = Pyro5.server._get_exposed_members(o)
        assert m["methods"] == {"oneway", "exposed"}
        assert m["attrs"] == {"prop1", "readonly_prop1"}
        assert m["oneway"] == {"oneway"}

    def testNotOnlyExposed(self):
        o = MyThingPartlyExposed("irmen")
        m = Pyro5.server._get_exposed_members(o, only_exposed=False)
        assert m["methods"] == {"classmethod", "staticmethod", "method", "__dunder__", "oneway", "exposed"}
        assert m["attrs"] == {"prop1", "readonly_prop1", "prop2"}
        assert m["oneway"] == {"oneway"}

    def testPartlyExposedSubclass(self):
        o = MyThingPartlyExposedSub("irmen")
        m = Pyro5.server._get_exposed_members(o)
        assert m["attrs"] == {"prop1", "readonly_prop1"}
        assert m["oneway"] == {"oneway"}
        assert m["methods"] == {"sub_exposed", "exposed", "oneway"}

    def testExposedSubclass(self):
        o = MyThingExposedSub("irmen")
        m = Pyro5.server._get_exposed_members(o)
        assert m["attrs"] == {"readonly_prop1", "prop1", "prop2"}
        assert m["oneway"] == {"oneway", "oneway2"}
        assert m["methods"] == {"classmethod", "staticmethod", "oneway", "__dunder__", "method", "exposed",
                                "oneway2", "sub_exposed", "sub_unexposed"}

    def testExposePrivateFails(self):
        with pytest.raises(AttributeError):
            class Test1(object):
                @Pyro5.server.expose
                def _private(self):
                    pass
        with pytest.raises(AttributeError):
            class Test3(object):
                @Pyro5.server.expose
                def __private(self):
                    pass
        with pytest.raises(AttributeError):
            @Pyro5.server.expose
            class _Test4(object):
                pass
        with pytest.raises(AttributeError):
            @Pyro5.server.expose
            class __Test5(object):
                pass

    def testExposeDunderOk(self):
        class Test1(object):
            @Pyro5.server.expose
            def __dunder__(self):
                pass
        assert Test1.__dunder__._pyroExposed
        @Pyro5.server.expose
        class Test2(object):
            def __dunder__(self):
                pass
        assert Test2._pyroExposed
        assert Test2.__dunder__._pyroExposed

    def testClassmethodExposeWrongOrderFail(self):
        with pytest.raises(AttributeError) as ax:
            class TestClass:
                @Pyro5.server.expose
                @classmethod
                def cmethod(cls):
                    pass
        assert "must be done after" in str(ax.value)
        with pytest.raises(AttributeError) as ax:
            class TestClass:
                @Pyro5.server.expose
                @staticmethod
                def smethod(cls):
                    pass
        assert "must be done after" in str(ax.value)

    def testClassmethodExposeCorrectOrderOkay(self):
        class TestClass:
            @classmethod
            @Pyro5.server.expose
            def cmethod(cls):
                pass
            @staticmethod
            @Pyro5.server.expose
            def smethod(cls):
                pass
        assert TestClass.cmethod._pyroExposed
        assert TestClass.smethod._pyroExposed

    def testGetExposedProperty(self):
        o = MyThingFullExposed("irmen")
        with pytest.raises(AttributeError):
            Pyro5.server._get_exposed_property_value(o, "name")
        with pytest.raises(AttributeError):
            Pyro5.server._get_exposed_property_value(o, "c_attr")
        with pytest.raises(AttributeError):
            Pyro5.server._get_exposed_property_value(o, "propvalue")
        with pytest.raises(AttributeError):
            Pyro5.server._get_exposed_property_value(o, "unexisting_attribute")
        assert Pyro5.server._get_exposed_property_value(o, "prop1") == 42
        assert Pyro5.server._get_exposed_property_value(o, "prop2") == 42

    def testGetExposedPropertyFromPartiallyExposed(self):
        o = MyThingPartlyExposed("irmen")
        with pytest.raises(AttributeError):
            Pyro5.server._get_exposed_property_value(o, "name")
        with pytest.raises(AttributeError):
            Pyro5.server._get_exposed_property_value(o, "c_attr")
        with pytest.raises(AttributeError):
            Pyro5.server._get_exposed_property_value(o, "propvalue")
        with pytest.raises(AttributeError):
            Pyro5.server._get_exposed_property_value(o, "unexisting_attribute")
        assert Pyro5.server._get_exposed_property_value(o, "prop1") == 42
        with pytest.raises(AttributeError):
            Pyro5.server._get_exposed_property_value(o, "prop2")

    def testSetExposedProperty(self):
        o = MyThingFullExposed("irmen")
        with pytest.raises(AttributeError):
            Pyro5.server._set_exposed_property_value(o, "name", "erorr")
        with pytest.raises(AttributeError):
            Pyro5.server._set_exposed_property_value(o, "unexisting_attribute", 42)
        with pytest.raises(AttributeError):
            Pyro5.server._set_exposed_property_value(o, "readonly_prop1", 42)
        with pytest.raises(AttributeError):
            Pyro5.server._set_exposed_property_value(o, "propvalue", 999)
        assert o.prop1 == 42
        assert o.prop2 == 42
        Pyro5.server._set_exposed_property_value(o, "prop1", 999)
        assert o.propvalue == 999
        Pyro5.server._set_exposed_property_value(o, "prop2", 8888)
        assert o.propvalue == 8888

    def testSetExposedPropertyFromPartiallyExposed(self):
        o = MyThingPartlyExposed("irmen")
        with pytest.raises(AttributeError):
            Pyro5.server._set_exposed_property_value(o, "name", "erorr")
        with pytest.raises(AttributeError):
            Pyro5.server._set_exposed_property_value(o, "unexisting_attribute", 42)
        with pytest.raises(AttributeError):
            Pyro5.server._set_exposed_property_value(o, "readonly_prop1", 42)
        with pytest.raises(AttributeError):
            Pyro5.server._set_exposed_property_value(o, "propvalue", 999)
        assert o.prop1 == 42
        assert o.prop2 == 42
        Pyro5.server._set_exposed_property_value(o, "prop1", 999)
        assert o.propvalue == 999
        with pytest.raises(AttributeError):
            Pyro5.server._set_exposed_property_value(o, "prop2", 8888)

    def testIsPrivateName(self):
        assert Pyro5.server.is_private_attribute("_")
        assert Pyro5.server.is_private_attribute("__")
        assert Pyro5.server.is_private_attribute("___")
        assert Pyro5.server.is_private_attribute("_p")
        assert Pyro5.server.is_private_attribute("_pp")
        assert Pyro5.server.is_private_attribute("_p_")
        assert Pyro5.server.is_private_attribute("_p__")
        assert Pyro5.server.is_private_attribute("__p")
        assert Pyro5.server.is_private_attribute("___p")
        assert not Pyro5.server.is_private_attribute("__dunder__")  # dunder methods should not be private except a list of exceptions as tested below
        assert Pyro5.server.is_private_attribute("__init__")
        assert Pyro5.server.is_private_attribute("__call__")
        assert Pyro5.server.is_private_attribute("__new__")
        assert Pyro5.server.is_private_attribute("__del__")
        assert Pyro5.server.is_private_attribute("__repr__")
        assert Pyro5.server.is_private_attribute("__str__")
        assert Pyro5.server.is_private_attribute("__format__")
        assert Pyro5.server.is_private_attribute("__nonzero__")
        assert Pyro5.server.is_private_attribute("__bool__")
        assert Pyro5.server.is_private_attribute("__coerce__")
        assert Pyro5.server.is_private_attribute("__cmp__")
        assert Pyro5.server.is_private_attribute("__eq__")
        assert Pyro5.server.is_private_attribute("__ne__")
        assert Pyro5.server.is_private_attribute("__lt__")
        assert Pyro5.server.is_private_attribute("__gt__")
        assert Pyro5.server.is_private_attribute("__le__")
        assert Pyro5.server.is_private_attribute("__ge__")
        assert Pyro5.server.is_private_attribute("__hash__")
        assert Pyro5.server.is_private_attribute("__dir__")
        assert Pyro5.server.is_private_attribute("__enter__")
        assert Pyro5.server.is_private_attribute("__exit__")
        assert Pyro5.server.is_private_attribute("__copy__")
        assert Pyro5.server.is_private_attribute("__deepcopy__")
        assert Pyro5.server.is_private_attribute("__sizeof__")
        assert Pyro5.server.is_private_attribute("__getattr__")
        assert Pyro5.server.is_private_attribute("__setattr__")
        assert Pyro5.server.is_private_attribute("__hasattr__")
        assert Pyro5.server.is_private_attribute("__delattr__")
        assert Pyro5.server.is_private_attribute("__getattribute__")
        assert Pyro5.server.is_private_attribute("__instancecheck__")
        assert Pyro5.server.is_private_attribute("__subclasscheck__")
        assert Pyro5.server.is_private_attribute("__subclasshook__")
        assert Pyro5.server.is_private_attribute("__getinitargs__")
        assert Pyro5.server.is_private_attribute("__getnewargs__")
        assert Pyro5.server.is_private_attribute("__getstate__")
        assert Pyro5.server.is_private_attribute("__setstate__")
        assert Pyro5.server.is_private_attribute("__reduce__")
        assert Pyro5.server.is_private_attribute("__reduce_ex__")

    def testResolveAttr(self):
        @Pyro5.server.expose
        class Exposed(object):
            def __init__(self, value):
                self.propvalue = value
                self.__value__ = value   # is not affected by the @expose

            def __str__(self):
                return "<%s>" % self.value

            def _p(self):
                return "should not be allowed"

            def __p(self):
                return "should not be allowed"

            def __p__(self):
                return "should be allowed (dunder)"

            @property
            def value(self):
                return self.propvalue

        class Unexposed(object):
            def __init__(self):
                self.value = 42

            def __value__(self):
                return self.value

        obj = Exposed("hello")
        obj.a = Exposed("a")
        obj.a.b = Exposed("b")
        obj.a.b.c = Exposed("c")
        obj.a._p = Exposed("p1")
        obj.a._p.q = Exposed("q1")
        obj.a.__p = Exposed("p2")
        obj.a.__p.q = Exposed("q2")
        obj.u = Unexposed()
        obj.u.v = Unexposed()
        # check the accessible attributes
        assert str(Pyro5.server._get_attribute(obj, "a")) == "<a>"
        dunder = str(Pyro5.server._get_attribute(obj, "__p__"))
        assert dunder.startswith("<bound method ")  # dunder is not private, part 1 of the check
        assert "Exposed.__p__ of" in dunder  # dunder is not private, part 2 of the check
        # check what should not be accessible
        with pytest.raises(AttributeError):
            Pyro5.server._get_attribute(obj, "value")
        with pytest.raises(AttributeError):
            Pyro5.server._get_attribute(obj, "propvalue")
        with pytest.raises(AttributeError):
            Pyro5.server._get_attribute(obj, "__value__")  # is not affected by the @expose
        with pytest.raises(AttributeError):
            Pyro5.server._get_attribute(obj, "_p")  # private
        with pytest.raises(AttributeError):
            Pyro5.server._get_attribute(obj, "__p")  # private
        with pytest.raises(AttributeError):
            Pyro5.server._get_attribute(obj, "a.b")
        with pytest.raises(AttributeError):
            Pyro5.server._get_attribute(obj, "a.b.c")
        with pytest.raises(AttributeError):
            Pyro5.server._get_attribute(obj, "a.b.c.d")
        with pytest.raises(AttributeError):
            Pyro5.server._get_attribute(obj, "a._p")
        with pytest.raises(AttributeError):
            Pyro5.server._get_attribute(obj, "a._p.q")
        with pytest.raises(AttributeError):
            Pyro5.server._get_attribute(obj, "a.__p.q")
        with pytest.raises(AttributeError):
            Pyro5.server._get_attribute(obj, "u")
        with pytest.raises(AttributeError):
            Pyro5.server._get_attribute(obj, "u.v")
        with pytest.raises(AttributeError):
            Pyro5.server._get_attribute(obj, "u.v.value")


class TestSimpleServe:
    class DaemonWrapper(Pyro5.server.Daemon):
        def requestLoop(self, *args):
            # override with empty method to fall out of the serve() call
            pass

    def testSimpleServeLegacy(self):
        with TestSimpleServe.DaemonWrapper() as d:
            o1 = MyThingPartlyExposed(1)
            o2 = MyThingPartlyExposed(2)
            objects = {o1: "test.o1", o2: None}
            Pyro5.server.serve(objects, daemon=d, use_ns=False, verbose=False)
            assert len(d.objectsById) == 3
            assert "test.o1" in d.objectsById
            assert o1 in d.objectsById.values()
            assert o2 in d.objectsById.values()

    def testSimpleServe(self):
        with TestSimpleServe.DaemonWrapper() as d:
            o1 = MyThingPartlyExposed(1)
            o2 = MyThingPartlyExposed(2)
            objects = {o1: "test.o1", o2: None}
            Pyro5.server.serve(objects, daemon=d, use_ns=False, verbose=False)
            assert len(d.objectsById) == 3
            assert "test.o1" in d.objectsById
            assert o1 in d.objectsById.values()
            assert o2 in d.objectsById.values()

    def testSimpleServeSameNamesLegacy(self):
        with TestSimpleServe.DaemonWrapper() as d:
            o1 = MyThingPartlyExposed(1)
            o2 = MyThingPartlyExposed(2)
            o3 = MyThingPartlyExposed(3)
            objects = {o1: "test.name", o2: "test.name", o3: "test.othername"}
            with pytest.raises(Pyro5.errors.DaemonError):
                Pyro5.server.serve(objects, daemon=d, use_ns=False, verbose=False)

    def testSimpleServeSameNames(self):
        with TestSimpleServe.DaemonWrapper() as d:
            o1 = MyThingPartlyExposed(1)
            o2 = MyThingPartlyExposed(2)
            o3 = MyThingPartlyExposed(3)
            objects = {o1: "test.name", o2: "test.name", o3: "test.othername"}
            with pytest.raises(Pyro5.errors.DaemonError):
                Pyro5.server.serve(objects, daemon=d, use_ns=False, verbose=False)


class TestExposeDecorator:
    # note: the bulk of the tests for the @expose decorator are found in the test_util module
    def testExposeInstancemodeDefault(self):
        @Pyro5.server.expose
        class TestClassOne:
            def method(self):
                pass
        class TestClassTwo:
            @Pyro5.server.expose
            def method(self):
                pass
        class TestClassThree:
            def method(self):
                pass
        with Pyro5.server.Daemon() as daemon:
            daemon.register(TestClassOne)
            daemon.register(TestClassTwo)
            daemon.register(TestClassThree)
            assert TestClassOne._pyroInstancing == ("session", None)
            assert TestClassTwo._pyroInstancing == ("session", None)
            assert TestClassThree._pyroInstancing == ("session", None)


class TestBehaviorDecorator:
    def testBehaviorInstancemodeInvalid(self):
        with pytest.raises(ValueError):
            @Pyro5.server.behavior(instance_mode="kaputt")
            class TestClass:
                def method(self):
                    pass

    def testBehaviorRequiresParams(self):
        with pytest.raises(SyntaxError) as x:
            @Pyro5.server.behavior
            class TestClass:
                def method(self):
                    pass
        assert "is missing argument" in str(x.value)

    def testBehaviorInstancecreatorInvalid(self):
        with pytest.raises(TypeError):
            @Pyro5.server.behavior(instance_creator=12345)
            class TestClass:
                def method(self):
                    pass

    def testBehaviorOnMethodInvalid(self):
        with pytest.raises(TypeError):
            class TestClass:
                @Pyro5.server.behavior(instance_mode="~invalidmode~")
                def method(self):
                    pass
        with pytest.raises(TypeError):
            class TestClass:
                @Pyro5.server.behavior(instance_mode="percall", instance_creator=float)
                def method(self):
                    pass
        with pytest.raises(TypeError):
            class TestClass:
                @Pyro5.server.behavior()
                def method(self):
                    pass

    def testBehaviorInstancing(self):
        @Pyro5.server.behavior(instance_mode="percall", instance_creator=float)
        class TestClass:
            def method(self):
                pass
        im, ic = TestClass._pyroInstancing
        assert im == "percall"
        assert ic is float

    def testBehaviorWithExposeKeepsCorrectValues(self):
        @Pyro5.server.behavior(instance_mode="percall", instance_creator=float)
        @Pyro5.server.expose
        class TestClass:
            pass
        im, ic = TestClass._pyroInstancing
        assert im == "percall"
        assert ic is float

        @Pyro5.server.expose
        @Pyro5.server.behavior(instance_mode="percall", instance_creator=float)
        class TestClass2:
            pass
        im, ic = TestClass2._pyroInstancing
        assert im == "percall"
        assert ic is float



import Pyro5.api
import uuid


# example: set a single correlation id on the context that should be passed along
Pyro5.api.current_context.correlation_id = uuid.uuid4()
print("correlation id set to:", Pyro5.api.current_context.correlation_id)


uri = input("Enter the URI of the server object: ")

print("\n------- get annotations via normal proxy and the call context... -----\n")

with Pyro5.api.Proxy(uri) as proxy:
    print("normal call")

    Pyro5.api.current_context.annotations = {"XYZZ": b"custom annotation from client (1)"}
    result = proxy.echo("hi there - new method of annotation access in client")
    print("Annotations in response were: ")
    for key, value in Pyro5.api.current_context.response_annotations.items():
        print("  ", key, "->", bytes(value))

    print("\noneway call")
    Pyro5.api.current_context.annotations = {"XYZZ": b"custom annotation from client (2)"}
    proxy.oneway("hi there ONEWAY - new method of annotation access in client")
    print("Annotations in response were: ")
    for key, value in Pyro5.api.current_context.response_annotations.items():
        print("  ", key, "->", bytes(value))
    print("(should be an empty result because oneway!)")


print("\nSee the console output on the server for more results.")



import sys
import os
import io
import pytest
from Pyro5 import config, errors


def crash(arg=100):
    pre1 = "black"
    pre2 = 999

    def nest(p1, p2):
        q = "white" + pre1
        x = pre2
        y = arg // 2
        p3 = p1 // p2
        return p3

    a = 10
    b = 0
    s = "hello"
    c = nest(a, b)
    return c


class TestErrors:
    def testFormatTracebackNormal(self):
        with pytest.raises(ZeroDivisionError) as x:
            crash()
        tb = "".join(errors.format_traceback(x.type, x.value, x.tb, detailed=False))
        assert "p3 = p1 // p2" in tb
        assert "ZeroDivisionError" in tb
        assert " a = 10" not in tb
        assert " s = 'whiteblack'" not in tb
        assert " pre2 = 999" not in tb
        assert " x = 999" not in  tb

    def testFormatTracebackDetail(self):
        with pytest.raises(ZeroDivisionError) as x:
            crash()
        tb = "".join(errors.format_traceback(x.type, x.value, x.tb, detailed=True))
        assert "p3 = p1 // p2" in tb
        assert "ZeroDivisionError" in tb
        assert " a = 10" in tb
        assert " q = 'whiteblack'" in tb
        assert " pre2 = 999" in tb
        assert " x = 999" in  tb

    def testPyroTraceback(self):
        try:
            crash()
        except ZeroDivisionError:
            pyro_tb = errors.format_traceback(detailed=True)
            assert " Extended stacktrace follows (most recent call last)\n" in pyro_tb
        try:
            crash("stringvalue")
        except TypeError as x:
            x._pyroTraceback = pyro_tb  # set the remote traceback info
            pyrotb = "".join(errors.get_pyro_traceback())
            assert "Remote traceback" in pyrotb
            assert "crash(\"stringvalue\")" in pyrotb
            assert "TypeError:" in pyrotb
            assert "ZeroDivisionError" in pyrotb
            del x._pyroTraceback
            pyrotb = "".join(errors.get_pyro_traceback())
            assert "Remote traceback" not in pyrotb
            assert "ZeroDivisionError" not in pyrotb
            assert "crash(\"stringvalue\")" in pyrotb
            assert "TypeError:" in pyrotb

    def testPyroTracebackArgs(self):
        try:
            crash()
        except ZeroDivisionError:
            ex_type, ex_value, ex_tb = sys.exc_info()
            tb1 = errors.get_pyro_traceback()
            tb2 = errors.get_pyro_traceback(ex_type, ex_value, ex_tb)
            assert tb2 == tb1
            tb1 = errors.format_traceback()
            tb2 = errors.format_traceback(ex_type, ex_value, ex_tb)
            assert tb2 == tb1
            tb2 = errors.format_traceback(detailed=True)
            assert tb1 != tb2

    def testExcepthook(self):
        # simply test the excepthook by calling it the way Python would
        try:
            crash()
        except ZeroDivisionError:
            pyro_tb = errors.format_traceback()
        with pytest.raises(TypeError) as x:
            crash("stringvalue")
        ex_type, ex_value, ex_tb = x.type, x.value, x.tb
        ex_value._pyroTraceback = pyro_tb  # set the remote traceback info
        oldstderr = sys.stderr
        try:
            sys.stderr = io.StringIO()
            errors.excepthook(ex_type, ex_value, ex_tb)
            output = sys.stderr.getvalue()
            assert "Remote traceback" in output
            assert "crash(\"stringvalue\")" in output
            assert "TypeError:" in output
            assert "ZeroDivisionError" in output
        finally:
            sys.stderr = oldstderr

    def clearEnv(self):
        if "PYRO_HOST" in os.environ:
            del os.environ["PYRO_HOST"]
        if "PYRO_NS_PORT" in os.environ:
            del os.environ["PYRO_NS_PORT"]
        if "PYRO_COMPRESSION" in os.environ:
            del os.environ["PYRO_COMPRESSION"]
        config.reset()

    def testConfig(self):
        self.clearEnv()
        try:
            assert config.NS_PORT == 9090
            assert config.HOST == "localhost"
            assert config.COMPRESSION == False
            os.environ["NS_PORT"] = "4444"
            config.reset()
            assert config.NS_PORT == 9090
            os.environ["PYRO_NS_PORT"] = "4444"
            os.environ["PYRO_HOST"] = "something.com"
            os.environ["PYRO_COMPRESSION"] = "OFF"
            config.reset()
            assert config.NS_PORT == 4444
            assert config.HOST == "something.com"
            assert config.COMPRESSION == False
        finally:
            self.clearEnv()
            assert config.NS_PORT == 9090
            assert config.HOST == "localhost"
            assert config.COMPRESSION == False

    def testConfigReset(self):
        try:
            config.reset()
            assert config.HOST == "localhost"
            config.HOST = "foobar"
            assert config.HOST == "foobar"
            config.reset()
            assert config.HOST == "localhost"
            os.environ["PYRO_HOST"] = "foobar"
            config.reset()
            assert config.HOST == "foobar"
            del os.environ["PYRO_HOST"]
            config.reset()
            assert config.HOST == "localhost"
        finally:
            self.clearEnv()
